# coding: utf-8

"""
FastAPI

No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

The version of the OpenAPI document: 0.1.0
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations

import json
import pprint
import re  # noqa: F401
from typing import Optional

from truefoundry.pydantic_v1 import BaseModel, StrictInt, StrictStr, conlist


class SearchRunsRequestDto(BaseModel):
    """
    SearchRunsRequestDto
    """

    experiment_ids: Optional[conlist(StrictStr)] = None
    filter: Optional[StrictStr] = None
    run_view_type: Optional[StrictStr] = "ACTIVE_ONLY"
    max_results: Optional[StrictInt] = 1000
    order_by: Optional[conlist(StrictStr)] = None
    page_token: Optional[StrictStr] = None
    offset: Optional[StrictInt] = None
    __properties = [
        "experiment_ids",
        "filter",
        "run_view_type",
        "max_results",
        "order_by",
        "page_token",
        "offset",
    ]

    class Config:
        """Pydantic configuration"""

        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> SearchRunsRequestDto:
        """Create an instance of SearchRunsRequestDto from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True, exclude={}, exclude_none=True)
        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> SearchRunsRequestDto:
        """Create an instance of SearchRunsRequestDto from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return SearchRunsRequestDto.parse_obj(obj)

        _obj = SearchRunsRequestDto.parse_obj(
            {
                "experiment_ids": obj.get("experiment_ids"),
                "filter": obj.get("filter"),
                "run_view_type": obj.get("run_view_type")
                if obj.get("run_view_type") is not None
                else "ACTIVE_ONLY",
                "max_results": obj.get("max_results")
                if obj.get("max_results") is not None
                else 1000,
                "order_by": obj.get("order_by"),
                "page_token": obj.get("page_token"),
                "offset": obj.get("offset"),
            }
        )
        return _obj



# coding: utf-8

"""
FastAPI

No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

The version of the OpenAPI document: 0.1.0
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations

import json
import pprint
import re  # noqa: F401
from typing import Optional

from truefoundry.ml.autogen.client.models.content1 import Content1
from truefoundry.pydantic_v1 import BaseModel, Field, StrictStr, validator


class SystemMessage(BaseModel):
    """
    +usage=System message +docs=Defines the structure of a system message, including role and content +label=System Message  # noqa: E501
    """

    role: StrictStr = Field(...)
    content: Content1 = Field(...)
    name: Optional[StrictStr] = Field(
        default=None, description="+label=Name +usage=Name of the system"
    )
    __properties = ["role", "content", "name"]

    @validator("role")
    def role_validate_enum(cls, value):
        """Validates the enum"""
        if value not in ("system",):
            raise ValueError("must be one of enum values ('system')")
        return value

    class Config:
        """Pydantic configuration"""

        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> SystemMessage:
        """Create an instance of SystemMessage from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True, exclude={}, exclude_none=True)
        # override the default output from truefoundry.pydantic_v1 by calling `to_dict()` of content
        if self.content:
            _dict["content"] = self.content.to_dict()
        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> SystemMessage:
        """Create an instance of SystemMessage from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return SystemMessage.parse_obj(obj)

        _obj = SystemMessage.parse_obj(
            {
                "role": obj.get("role"),
                "content": Content1.from_dict(obj.get("content"))
                if obj.get("content") is not None
                else None,
                "name": obj.get("name"),
            }
        )
        return _obj



import json
import logging
import os
import posixpath
from pathlib import Path, PureWindowsPath
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

from truefoundry.ml.exceptions import MlFoundryException
from truefoundry.ml.log_types.artifacts.constants import DESCRIPTION_MAX_LENGTH

logger = logging.getLogger(__name__)


def to_unix_path(path):
    path = os.path.normpath(path)
    if os.path.sep == "\\":
        path = PureWindowsPath(path).as_posix()
    return path


def _copy_tree(
    root_dir: str, src_path: str, dest_path: str, dest_to_src: Dict[str, str]
):
    os.makedirs(dest_path, exist_ok=True)
    for item in os.listdir(src_path):
        src = os.path.join(src_path, item)
        dest = os.path.join(dest_path, item)
        if os.path.isdir(src):
            _copy_tree(
                root_dir=root_dir,
                src_path=src,
                dest_path=dest,
                dest_to_src=dest_to_src,
            )
        else:
            dest_to_src[dest] = src


def is_destination_path_dirlike(dest_path) -> bool:
    if not dest_path:
        return True

    if dest_path.endswith(os.sep) or dest_path.endswith(posixpath.sep):
        return True

    if os.path.exists(dest_path) and os.path.isdir(dest_path):
        return True

    return False


def get_single_file_path_if_only_one_in_directory(path: str) -> Optional[str]:
    """
    Get the filename from a path, or return a filename from a directory if a single file exists.
    Args:
        path: The file or folder path.
    Returns:
        Optional[str]: The filename or None if no files are found or multiple files are found.
    """
    # If it's already a file, return it as-is
    if os.path.isfile(path):
        return path

    # If it's a directory, check if it contains a single file
    if is_destination_path_dirlike(path):
        all_files: List[str] = []
        for root, _, files in os.walk(path):
            # Collect all files found in any subdirectory
            all_files.extend(os.path.join(root, f) for f in files)
            # If more than one file is found, stop early
            if len(all_files) > 1:
                return None

        # If only one file is found, return it
        if len(all_files) == 1:
            return all_files[0]

    return None  # No file found or Multiple files found


def _copy_additional_files(
    root_dir: str,
    files_dir: str,  # relative to root dir e.g. "files/"
    model_dir: Optional[str],  # relative to files_dir e.g "model/"
    additional_files: Sequence[Tuple[Union[str, Path], Optional[str]]],
    ignore_model_dir_dest_conflict: bool = False,
    existing_dest_to_src_map: Optional[Dict[str, str]] = None,
) -> Dict[str, str]:
    """
    File copying examples:
        # non ambiguous
        # a.txt -> /tmp/                            result /tmp/a.txt
        # a.txt -> /tmp/a/                          result /tmp/a/a.txt
        # a.txt -> /tmp/a/b/c/d.txt                 result /tmp/a/b/c/d.txt
        # .gitignore -> /tmp/.gitignore             result /tmp/.gitignore

        # ambiguous but destination directory exists
        # a.txt -> /tmp                             result /tmp/a.txt
        # a.txt -> /tmp/a (and /tmp/a/ exists)      result /tmp/a/a.txt

        # ambiguous - when the destination can't be reliably distinguished as a directory
        # a -> /tmp/a                                result /tmp/a
        # a -> /tmp/b                                result /tmp/b
        # a -> /tmp/a.txt                            result /tmp/a.txt
        # .gitignore -> /tmp/.gitinclude             result /tmp/.gitinclude
        # a.txt -> /tmp/a                            result /tmp/a
    """
    dest_to_src = existing_dest_to_src_map or {}
    for src_path, dest_path in additional_files:
        src_path = str(src_path)
        if not os.path.exists(src_path):
            raise MlFoundryException(
                f"Source path {src_path!r} in `additional_files` does not exist."
            )
        dest_path = dest_path or ""
        normalized_path = os.path.normpath(dest_path)
        if dest_path.endswith(os.sep) or dest_path.endswith(posixpath.sep):
            normalized_path += os.sep
        dest_path = normalized_path.lstrip(os.sep)

        if (
            model_dir
            and dest_path.startswith(model_dir)
            and not ignore_model_dir_dest_conflict
        ):
            logger.warning(
                f"Destination path {dest_path!r} in `additional_files` conflicts with "
                f"reserved path {model_dir!r}/ which is being used to store the model. "
                f"This might cause errors"
            )

        files_abs_dir = os.path.join(root_dir, files_dir)
        dest_abs_path = os.path.join(files_abs_dir, dest_path)

        if os.path.isfile(src_path):
            _src = src_path
            if is_destination_path_dirlike(dest_abs_path):
                os.makedirs(dest_abs_path, exist_ok=True)
                dest_abs_path = os.path.join(dest_abs_path, os.path.basename(_src))
            else:
                os.makedirs(os.path.dirname(dest_abs_path), exist_ok=True)
            _dst = os.path.relpath(dest_abs_path, files_abs_dir)
            logger.info(f"Adding file {_src} as /{_dst}")
            dest_to_src[dest_abs_path] = src_path
        elif os.path.isdir(src_path):
            os.makedirs(dest_abs_path, exist_ok=True)
            _src = src_path.rstrip("/")
            _dst = os.path.relpath(dest_abs_path, files_abs_dir).rstrip("/")
            logger.info(f"Adding contents of {_src}/ to /{_dst}/")
            _copy_tree(
                root_dir=root_dir,
                src_path=src_path,
                dest_path=dest_abs_path,
                dest_to_src=dest_to_src,
            )

    return dest_to_src


def _make_dest_to_src_map_from_dir(root_dir: str) -> Dict[str, str]:
    dest_to_src_map = {}
    for root, _, files in os.walk(root_dir):
        for file in files:
            src = os.path.join(root, file)
            dest = src
            dest_to_src_map[dest] = src
    return dest_to_src_map


def _get_src_dest_pairs(
    root_dir: str,
    dest_to_src_map: Dict[str, str],
) -> Sequence[Tuple[str, str]]:
    src_dest_pairs = [
        (src_path, to_unix_path(os.path.relpath(dest_abs_path, root_dir)))
        for dest_abs_path, src_path in dest_to_src_map.items()
    ]
    return src_dest_pairs


def _validate_description(description: Optional[str]):
    if description is not None:
        if not isinstance(description, str):
            raise MlFoundryException(
                "`description` must be either `None` or type `str`"
            )
        if len(description) > DESCRIPTION_MAX_LENGTH:
            raise MlFoundryException(
                f"`description` cannot be longer than {DESCRIPTION_MAX_LENGTH} characters"
            )


def _validate_artifact_metadata(metadata: Dict[str, Any]):
    if not isinstance(metadata, dict):
        raise MlFoundryException("`metadata` must be json serializable dict")
    try:
        json.dumps(metadata)
    except ValueError as ve:
        raise MlFoundryException("`metadata` must be json serializable dict") from ve


def calculate_total_size(
    paths: Sequence[str],
):
    """
    Tells about the size of the artifact

    Args:
        directory (str): directory path

    Returns:
        total size of the artifact
    """
    return sum(os.stat(os.path.realpath(file_path)).st_size for file_path in paths)



from typing import List, Tuple

import rich_click as click

from truefoundry.cli.console import console
from truefoundry.cli.const import GROUP_CLS
from truefoundry.cli.util import handle_exception_wrapper
from truefoundry.deploy.lib.clients.servicefoundry_client import (
    ServiceFoundryServiceClient,
)
from truefoundry.deploy.lib.dao import apply as apply_lib
from truefoundry.deploy.lib.messages import PROMPT_APPLYING_MANIFEST
from truefoundry.deploy.lib.model.entity import ApplyResult


@click.group(
    name="apply",
    cls=GROUP_CLS,
    invoke_without_command=True,
    help="Create resources by applying manifest locally from TrueFoundry spec",
    context_settings={"ignore_unknown_options": True, "allow_interspersed_args": True},
)
@click.option(
    "-f",
    "--file",
    "files",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    help="Path to yaml manifest file (You can apply multiple files at once by providing multiple -f options)",
    show_default=True,
    required=True,
    multiple=True,
)
@click.option(
    "--dry-run",
    "--dry_run",
    is_flag=True,
    show_default=True,
    help="Simulate the process without actually applying the manifest",
)
@handle_exception_wrapper
def apply_command(files: Tuple[str, ...], dry_run: bool = False):
    apply_results: List[ApplyResult] = []
    client = ServiceFoundryServiceClient()
    for file in files:
        with console.status(PROMPT_APPLYING_MANIFEST.format(file), spinner="dots"):
            for apply_result in apply_lib.apply_manifest_file(file, client, dry_run):
                if apply_result.success:
                    console.print(f"[green]\u2714 {apply_result.message}[/]")
                else:
                    console.print(f"[red]\u2718 {apply_result.message}[/]")

                apply_results.append(apply_result)

    if not all(apply_result.success for apply_result in apply_results):
        raise Exception("Failed to apply one or more manifests")


def get_apply_command():
    return apply_command



import functools
import sys
import zipfile
from typing import Dict

import rich_click as click
from packaging.version import parse as parse_version
from requests.exceptions import ConnectionError, Timeout
from rich.padding import Padding
from rich.panel import Panel
from rich.table import Table

from truefoundry.cli.console import console
from truefoundry.common.exceptions import HttpRequestException
from truefoundry.common.utils import is_debug_env_set


def setup_rich_click():
    click.rich_click.STYLE_ERRORS_SUGGESTION = "blue italic"
    click.rich_click.SHOW_ARGUMENTS = True
    click.rich_click.USE_RICH_MARKUP = True
    click.rich_click.STYLE_HELPTEXT = ""


def handle_exception(exception):
    if is_debug_env_set():
        console.print_exception(show_locals=True)
    if isinstance(exception, HttpRequestException):
        print_dict_as_table_panel(
            {"Status Code": str(exception.status_code), "Error": exception.message},
            title="Command Failed",
            border_style="red",
        )
    elif isinstance(exception, Timeout):
        loc = ""
        if exception.request:
            loc = f" at {exception.request.url}"
        print_dict_as_table_panel(
            {"Error": f"Request to TrueFoundry{loc} timed out."},
            title="Command Failed",
            border_style="red",
        )
    elif isinstance(exception, ConnectionError):
        loc = ""
        if exception.request:
            loc = f" at {exception.request.url}"
        print_dict_as_table_panel(
            {
                "Error": f"Couldn't connect to TrueFoundry{loc}. Please make sure that the provided `--host` is correct."
            },
            title="Command Failed",
            border_style="red",
        )
    else:
        print_dict_as_table_panel(
            {"Error": str(exception)},
            title="Command Failed",
            border_style="red",
        )


def handle_exception_wrapper(fn):
    @functools.wraps(fn)
    def inner(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except Exception as e:
            handle_exception(e)
            sys.exit(1)

    return inner


def print_error(message):
    text = Padding(message, (0, 1))
    console.print(
        Panel(
            text,
            border_style="red",
            title="Command failed",
            title_align="left",
            width=click.rich_click.MAX_WIDTH,
        )
    )


def print_message(message):
    text = Padding(message, (0, 1))
    console.print(
        Panel(
            text,
            border_style="cyan",
            title="Success",
            title_align="left",
            width=click.rich_click.MAX_WIDTH,
        )
    )


def print_dict_as_table_panel(
    dct: Dict[str, str],
    title: str,
    border_style: str = "cyan",
    key_color: str = "cyan",
):
    table = Table(show_header=False, box=None)
    table.add_column("Key", style=f"bold {key_color}", width=15)
    table.add_column("Value")
    for key, value in dct.items():
        table.add_row(key, value)
    console.print(
        Panel(
            table,
            border_style=border_style,
            title=title,
            title_align="left",
            width=click.rich_click.MAX_WIDTH,
        )
    )


def unzip_package(path_to_package, destination):
    with zipfile.ZipFile(path_to_package, "r") as zip_ref:
        zip_ref.extractall(destination)


def _prompt_if_no_value_and_supported(prompt: str, hide_input: bool = True):
    import click as _click

    kwargs = {}
    if parse_version(_click.__version__).major >= 8:
        kwargs = {"prompt": prompt, "hide_input": hide_input, "prompt_required": False}

    return kwargs



# coding: utf-8

"""
FastAPI

No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

The version of the OpenAPI document: 0.1.0
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations

import json
import pprint
import re  # noqa: F401
from typing import Optional

from truefoundry.ml.autogen.client.models.artifact_version_dto import ArtifactVersionDto
from truefoundry.pydantic_v1 import BaseModel, Field, StrictInt, StrictStr, conlist


class ListArtifactVersionsResponseDto(BaseModel):
    """
    ListArtifactVersionsResponseDto
    """

    artifact_versions: conlist(ArtifactVersionDto) = Field(...)
    next_page_token: Optional[StrictStr] = None
    total: Optional[StrictInt] = None
    __properties = ["artifact_versions", "next_page_token", "total"]

    class Config:
        """Pydantic configuration"""

        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> ListArtifactVersionsResponseDto:
        """Create an instance of ListArtifactVersionsResponseDto from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True, exclude={}, exclude_none=True)
        # override the default output from truefoundry.pydantic_v1 by calling `to_dict()` of each item in artifact_versions (list)
        _items = []
        if self.artifact_versions:
            for _item in self.artifact_versions:
                if _item:
                    _items.append(_item.to_dict())
            _dict["artifact_versions"] = _items
        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> ListArtifactVersionsResponseDto:
        """Create an instance of ListArtifactVersionsResponseDto from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return ListArtifactVersionsResponseDto.parse_obj(obj)

        _obj = ListArtifactVersionsResponseDto.parse_obj(
            {
                "artifact_versions": [
                    ArtifactVersionDto.from_dict(_item)
                    for _item in obj.get("artifact_versions")
                ]
                if obj.get("artifact_versions") is not None
                else None,
                "next_page_token": obj.get("next_page_token"),
                "total": obj.get("total"),
            }
        )
        return _obj



# coding: utf-8

"""
FastAPI

No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

The version of the OpenAPI document: 0.1.0
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations

import json
import pprint
import re  # noqa: F401
from typing import Any, Dict, Optional

from truefoundry.ml.autogen.client.models.framework import Framework
from truefoundry.ml.autogen.client.models.model_version_environment import (
    ModelVersionEnvironment,
)
from truefoundry.ml.autogen.client.models.source1 import Source1
from truefoundry.pydantic_v1 import (
    BaseModel,
    Field,
    StrictStr,
    conint,
    constr,
    validator,
)


class ModelVersionManifest(BaseModel):
    """
    Model Version artifact.  # noqa: E501
    """

    description: Optional[constr(strict=True, max_length=512)] = Field(
        default=None,
        description="+label=Description +usage=Description of the artifact or model version +docs=Description of the artifact or model version",
    )
    metadata: Dict[str, Any] = Field(
        default=...,
        description="+label=Metadata +docs=Metadata for the artifact or model version +usage=Metadata for the artifact or model version +uiType=JsonInput",
    )
    version_alias: Optional[constr(strict=True, max_length=128)] = Field(
        default=None,
        description="+label=Version Alias +usage=The version alias for artifact or model version which should start with 'v' followed by alphanumeric and it can include '.' and '-' in between (e.g. v1.0.0, v-prod, v-dev, etc) +docs=The version alias for artifact or model version which should start with 'v' followed by alphanumeric and it can include '.' and '-' in between (e.g. v1.0.0, v-prod, v-dev, etc) +message=The version alias should start with 'v' followed by alphanumeric and it can include '.' and '-' in between (e.g. v1.0.0, v-prod, v-dev, etc)",
    )
    type: Optional[StrictStr] = "model-version"
    source: Source1 = Field(...)
    framework: Optional[Framework] = None
    environment: Optional[ModelVersionEnvironment] = None
    step: Optional[conint(strict=True, ge=0)] = Field(
        default=0, description="+label=Step"
    )
    __properties = [
        "description",
        "metadata",
        "version_alias",
        "type",
        "source",
        "framework",
        "environment",
        "step",
    ]

    @validator("version_alias")
    def version_alias_validate_regular_expression(cls, value):
        """Validates the regular expression"""
        if value is None:
            return value

        if not re.match(r"^v[a-zA-Z0-9.-]*([a-zA-Z0-9]+)$", value):
            raise ValueError(
                r"must validate the regular expression /^v[a-zA-Z0-9.-]*([a-zA-Z0-9]+)$/"
            )
        return value

    @validator("type")
    def type_validate_enum(cls, value):
        """Validates the enum"""
        if value is None:
            return value

        if value not in ("model-version",):
            raise ValueError("must be one of enum values ('model-version')")
        return value

    class Config:
        """Pydantic configuration"""

        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> ModelVersionManifest:
        """Create an instance of ModelVersionManifest from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True, exclude={}, exclude_none=True)
        # override the default output from truefoundry.pydantic_v1 by calling `to_dict()` of source
        if self.source:
            _dict["source"] = self.source.to_dict()
        # override the default output from truefoundry.pydantic_v1 by calling `to_dict()` of framework
        if self.framework:
            _dict["framework"] = self.framework.to_dict()
        # override the default output from truefoundry.pydantic_v1 by calling `to_dict()` of environment
        if self.environment:
            _dict["environment"] = self.environment.to_dict()
        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> ModelVersionManifest:
        """Create an instance of ModelVersionManifest from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return ModelVersionManifest.parse_obj(obj)

        _obj = ModelVersionManifest.parse_obj(
            {
                "description": obj.get("description"),
                "metadata": obj.get("metadata"),
                "version_alias": obj.get("version_alias"),
                "type": obj.get("type")
                if obj.get("type") is not None
                else "model-version",
                "source": Source1.from_dict(obj.get("source"))
                if obj.get("source") is not None
                else None,
                "framework": Framework.from_dict(obj.get("framework"))
                if obj.get("framework") is not None
                else None,
                "environment": ModelVersionEnvironment.from_dict(obj.get("environment"))
                if obj.get("environment") is not None
                else None,
                "step": obj.get("step") if obj.get("step") is not None else 0,
            }
        )
        return _obj



import enum
import os
from pathlib import Path
from typing import Optional

from truefoundry.pydantic_v1 import BaseSettings, Field, SecretStr

TFY_CONFIG_DIR = Path.home() / ".truefoundry"
CREDENTIAL_FILEPATH = TFY_CONFIG_DIR / "credentials.json"

# These keys are kept separately because we use them in error messages and some checks
TFY_HOST_ENV_KEY = "TFY_HOST"
TFY_API_KEY_ENV_KEY = "TFY_API_KEY"
TFY_DEBUG_ENV_KEY = "TFY_DEBUG"
TFY_INTERNAL_ENV_KEY = "TFY_INTERNAL"

TFY_INTERNAL_SIGNED_URL_SERVER_HOST_ENV_KEY = "TFY_INTERNAL_SIGNED_URL_SERVER_HOST"
TFY_INTERNAL_SIGNED_URL_SERVER_TOKEN_ENV_KEY = "TFY_INTERNAL_SIGNED_URL_SERVER_TOKEN"


class PythonPackageManager(str, enum.Enum):
    PIP = "pip"
    UV = "uv"


class TrueFoundrySdkEnv(BaseSettings):
    # Note: Every field in this class should have a default value
    # Never expect the user to set these values
    TFY_HOST: Optional[str] = Field(default=None, env=TFY_HOST_ENV_KEY)
    TFY_API_KEY: Optional[SecretStr] = Field(default=None, env=TFY_API_KEY_ENV_KEY)

    # Internal Signed URL Server
    TFY_INTERNAL_SIGNED_URL_SERVER_HOST: Optional[str] = Field(
        default=None, env=TFY_INTERNAL_SIGNED_URL_SERVER_HOST_ENV_KEY
    )
    TFY_INTERNAL_SIGNED_URL_SERVER_TOKEN: Optional[str] = Field(
        default=None, env=TFY_INTERNAL_SIGNED_URL_SERVER_TOKEN_ENV_KEY
    )
    TFY_INTERNAL_SIGNED_URL_SERVER_MAX_TIMEOUT: int = 5  # default: 5 seconds
    TFY_INTERNAL_SIGNED_URL_SERVER_DEFAULT_TTL: int = 3600  # default: 1 hour
    TFY_INTERNAL_MULTIPART_UPLOAD_FINALIZE_SIGNED_URL_TIMEOUT: int = (
        24 * 60 * 60
    )  # default: 24 hour
    TFY_INTERNAL_SIGNED_URL_REQUEST_TIMEOUT: int = 3600  # default: 1 hour
    TFY_INTERNAL_SIGNED_URL_CLIENT_LOG_LEVEL: str = "WARNING"

    # Artifacts
    TFY_ARTIFACTS_DOWNLOAD_CHUNK_SIZE_BYTES: int = 100 * 1000 * 1000
    TFY_ARTIFACTS_DOWNLOAD_MAX_WORKERS: int = max(min(32, (os.cpu_count() or 2) * 2), 4)
    TFY_ARTIFACTS_UPLOAD_MAX_WORKERS: int = max(min(32, (os.cpu_count() or 2) * 2), 4)
    TFY_ARTIFACTS_DISABLE_MULTIPART_UPLOAD: bool = False
    TFY_ARTIFACTS_DOWNLOAD_FSYNC_CHUNKS: bool = False

    # Fo customizing the python image used for building via PythonBuild
    TFY_PYTHONBUILD_PYTHON_IMAGE_REPO: str = "public.ecr.aws/docker/library/python"

    # For local development, this enables futher configuration via _TFYServersConfig
    TFY_CLI_LOCAL_DEV_MODE: bool = False

    TFY_PYTHON_BUILD_PACKAGE_MANAGER: PythonPackageManager = PythonPackageManager.PIP
    TFY_PYTHON_BUILD_UV_IMAGE_URI: str = "ghcr.io/astral-sh/uv:latest"


ENV_VARS = TrueFoundrySdkEnv()
API_SERVER_RELATIVE_PATH = "api/svc"
MLFOUNDRY_SERVER_RELATIVE_PATH = "api/ml"
VERSION_PREFIX = "v1"
SERVICEFOUNDRY_CLIENT_MAX_RETRIES = 2



import os
import posixpath
import sys
import tempfile
import uuid
from concurrent.futures import FIRST_EXCEPTION, Future, ThreadPoolExecutor, wait
from shutil import rmtree
from threading import Event
from typing import (
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)
from urllib.parse import unquote
from urllib.request import pathname2url

from rich.console import _is_jupyter
from rich.progress import (
    BarColumn,
    DownloadColumn,
    Progress,
    TimeElapsedColumn,
    TimeRemainingColumn,
    TransferSpeedColumn,
)
from tqdm.utils import CallbackIOWrapper

from truefoundry.common.constants import ENV_VARS
from truefoundry.common.request_utils import (
    augmented_raise_for_status,
    cloud_storage_http_request,
)
from truefoundry.common.storage_provider_utils import (
    MultiPartUpload,
    _FileMultiPartInfo,
    azure_multi_part_upload,
    decide_file_parts,
    s3_compatible_multipart_upload,
)
from truefoundry.ml.autogen.client import (  # type: ignore[attr-defined]
    ApiClient,
    CreateMultiPartUploadForDatasetRequestDto,
    CreateMultiPartUploadRequestDto,
    FileInfoDto,
    GetSignedURLForDatasetWriteRequestDto,
    GetSignedURLsForArtifactVersionReadRequestDto,
    GetSignedURLsForArtifactVersionWriteRequestDto,
    GetSignedURLsForDatasetReadRequestDto,
    ListFilesForArtifactVersionRequestDto,
    ListFilesForArtifactVersionsResponseDto,
    ListFilesForDatasetRequestDto,
    ListFilesForDatasetResponseDto,
    MlfoundryArtifactsApi,
    MultiPartUploadDto,
    MultiPartUploadResponseDto,
    MultiPartUploadStorageProvider,
    RunArtifactsApi,
    SignedURLDto,
)
from truefoundry.ml.exceptions import MlFoundryException
from truefoundry.ml.logger import logger
from truefoundry.ml.session import _get_api_client
from truefoundry.pydantic_v1 import BaseModel, root_validator

_LIST_FILES_PAGE_SIZE = 500
_GENERATE_SIGNED_URL_BATCH_SIZE = 50
DEFAULT_PRESIGNED_URL_EXPIRY_TIME = 3600


def _can_display_progress(user_choice: Optional[bool] = None) -> bool:
    if user_choice is False:
        return False

    if sys.stdout.isatty():
        return True
    elif _is_jupyter():
        try:
            from IPython.display import display  # noqa: F401
            from ipywidgets import Output  # noqa: F401

            return True
        except ImportError:
            logger.warning(
                "Detected Jupyter Environment. Install `ipywidgets` to display live progress bars.",
            )
    if user_choice is True:
        logger.warning(
            "`progress` argument is set to True but did not detect tty "
            "or jupyter environment with ipywidgets installed. "
            "Progress bars may not be displayed. "
        )
        return True
    return False


def relative_path_to_artifact_path(path):
    if os.path == posixpath:
        return path
    if os.path.abspath(path) == path:
        raise Exception("This method only works with relative paths.")
    return unquote(pathname2url(path))


def bad_path_message(name):
    return (
        "Names may be treated as files in certain cases, and must not resolve to other names"
        " when treated as such. This name would resolve to '%s'"
    ) % posixpath.normpath(name)


def path_not_unique(name):
    norm = posixpath.normpath(name)
    return norm != name or norm == "." or norm.startswith("..") or norm.startswith("/")


def verify_artifact_path(artifact_path):
    if artifact_path and path_not_unique(artifact_path):
        raise MlFoundryException(
            f"Invalid artifact path: {artifact_path!r}. {bad_path_message(artifact_path)!r}"
        )


def _signed_url_upload_file(
    signed_url: SignedURLDto,
    local_file: str,
    progress_bar: Progress,
    abort_event: Optional[Event] = None,
):
    if os.stat(local_file).st_size == 0:
        with cloud_storage_http_request(
            method="put",
            url=signed_url.signed_url,
            data="",
            exception_class=MlFoundryException,  # type: ignore
        ) as response:
            augmented_raise_for_status(response, exception_class=MlFoundryException)  # type: ignore
        return

    task_progress_bar = progress_bar.add_task(
        f"[green]Uploading {local_file}:", start=True
    )

    def callback(length):
        progress_bar.update(
            task_progress_bar, advance=length, total=os.stat(local_file).st_size
        )
        if abort_event and abort_event.is_set():
            raise Exception("aborting upload")

    with open(local_file, "rb") as file:
        # NOTE: Azure Put Blob does not support Transfer Encoding header.
        wrapped_file = CallbackIOWrapper(callback, file, "read")
        with cloud_storage_http_request(
            method="put",
            url=signed_url.signed_url,
            data=wrapped_file,
            exception_class=MlFoundryException,  # type: ignore
        ) as response:
            augmented_raise_for_status(response, exception_class=MlFoundryException)  # type: ignore


def _download_file_using_http_uri(
    http_uri,
    download_path,
    chunk_size=ENV_VARS.TFY_ARTIFACTS_DOWNLOAD_CHUNK_SIZE_BYTES,
    callback: Optional[Callable[[int, int], Any]] = None,
):
    """
    Downloads a file specified using the `http_uri` to a local `download_path`. This function
    uses a `chunk_size` to ensure an OOM error is not raised a large file is downloaded.
    Note : This function is meant to download files using presigned urls from various cloud
            providers.
    """
    with cloud_storage_http_request(
        method="get",
        url=http_uri,
        stream=True,
        exception_class=MlFoundryException,  # type: ignore
    ) as response:
        augmented_raise_for_status(response, exception_class=MlFoundryException)  # type: ignore
        file_size = int(response.headers.get("Content-Length", 0))
        with open(download_path, "wb") as output_file:
            for chunk in response.iter_content(chunk_size=chunk_size):
                if callback:
                    callback(len(chunk), file_size)
                if not chunk:
                    break
                output_file.write(chunk)
                if ENV_VARS.TFY_ARTIFACTS_DOWNLOAD_FSYNC_CHUNKS:
                    output_file.flush()
                    os.fsync(output_file.fileno())


def _any_future_has_failed(futures) -> bool:
    return any(
        future.done() and not future.cancelled() and future.exception() is not None
        for future in futures
    )


class ArtifactIdentifier(BaseModel):
    artifact_version_id: Optional[uuid.UUID] = None
    dataset_fqn: Optional[str] = None

    @root_validator
    def _check_identifier_type(cls, values: Dict[str, Any]):
        if not values.get("artifact_version_id", False) and not values.get(
            "dataset_fqn", False
        ):
            raise MlFoundryException(
                "One of the version_id or dataset_fqn should be passed"
            )
        if values.get("artifact_version_id", False) and values.get(
            "dataset_fqn", False
        ):
            raise MlFoundryException(
                "Exactly one of version_id or dataset_fqn should be passed"
            )
        return values


class MlFoundryArtifactsRepository:
    def __init__(
        self,
        artifact_identifier: ArtifactIdentifier,
        api_client: Optional[ApiClient] = None,
    ):
        self.artifact_identifier = artifact_identifier
        self._api_client = api_client or _get_api_client()
        self._run_artifacts_api = RunArtifactsApi(api_client=self._api_client)
        self._mlfoundry_artifacts_api = MlfoundryArtifactsApi(
            api_client=self._api_client
        )

    # TODO (chiragjn): Refactor these methods - if else is very inconvenient
    def get_signed_urls_for_read(
        self,
        artifact_identifier: ArtifactIdentifier,
        paths,
    ) -> List[SignedURLDto]:
        if artifact_identifier.artifact_version_id:
            signed_urls_response = self._mlfoundry_artifacts_api.get_signed_urls_for_read_post(
                get_signed_urls_for_artifact_version_read_request_dto=GetSignedURLsForArtifactVersionReadRequestDto(
                    id=str(artifact_identifier.artifact_version_id), paths=paths
                )
            )
            signed_urls = signed_urls_response.signed_urls
        elif artifact_identifier.dataset_fqn:
            signed_urls_dataset_response = self._mlfoundry_artifacts_api.get_signed_urls_dataset_read_post(
                get_signed_urls_for_dataset_read_request_dto=GetSignedURLsForDatasetReadRequestDto(
                    dataset_fqn=artifact_identifier.dataset_fqn, paths=paths
                )
            )
            signed_urls = signed_urls_dataset_response.signed_urls
        else:
            raise ValueError(
                "Invalid artifact type - both `artifact_version_id` and `dataset_fqn` both are None"
            )
        return signed_urls

    def get_signed_urls_for_write(
        self,
        artifact_identifier: ArtifactIdentifier,
        paths: List[str],
    ) -> List[SignedURLDto]:
        if artifact_identifier.artifact_version_id:
            signed_urls_response = self._mlfoundry_artifacts_api.get_signed_urls_for_write_post(
                get_signed_urls_for_artifact_version_write_request_dto=GetSignedURLsForArtifactVersionWriteRequestDto(
                    id=str(artifact_identifier.artifact_version_id), paths=paths
                )
            )
            signed_urls = signed_urls_response.signed_urls
        elif artifact_identifier.dataset_fqn:
            signed_urls_dataset_response = self._mlfoundry_artifacts_api.get_signed_urls_for_dataset_write_post(
                get_signed_url_for_dataset_write_request_dto=GetSignedURLForDatasetWriteRequestDto(
                    dataset_fqn=artifact_identifier.dataset_fqn, paths=paths
                )
            )
            signed_urls = signed_urls_dataset_response.signed_urls
        else:
            raise ValueError(
                "Invalid artifact type - both `artifact_version_id` and `dataset_fqn` both are None"
            )
        return signed_urls

    def _normal_upload(
        self,
        local_file: str,
        artifact_path: str,
        signed_url: Optional[SignedURLDto],
        progress_bar: Progress,
        abort_event: Optional[Event] = None,
    ):
        if not signed_url:
            signed_url = self.get_signed_urls_for_write(
                artifact_identifier=self.artifact_identifier, paths=[artifact_path]
            )[0]

        if progress_bar.disable:
            logger.info(
                "Uploading %s to %s",
                local_file,
                artifact_path,
            )

        _signed_url_upload_file(
            signed_url=signed_url,
            local_file=local_file,
            abort_event=abort_event,
            progress_bar=progress_bar,
        )
        logger.debug("Uploaded %s to %s", local_file, artifact_path)

    def _create_multipart_upload_for_identifier(
        self,
        artifact_identifier: ArtifactIdentifier,
        path,
        num_parts,
    ) -> MultiPartUploadDto:
        if artifact_identifier.artifact_version_id:
            create_multipart_response: MultiPartUploadResponseDto = self._mlfoundry_artifacts_api.create_multi_part_upload_post(
                create_multi_part_upload_request_dto=CreateMultiPartUploadRequestDto(
                    artifact_version_id=str(artifact_identifier.artifact_version_id),
                    path=path,
                    num_parts=num_parts,
                )
            )
            multipart_upload = create_multipart_response.multipart_upload
        elif artifact_identifier.dataset_fqn:
            create_multipart_for_dataset_response = self._mlfoundry_artifacts_api.create_multipart_upload_for_dataset_post(
                create_multi_part_upload_for_dataset_request_dto=CreateMultiPartUploadForDatasetRequestDto(
                    dataset_fqn=artifact_identifier.dataset_fqn,
                    path=path,
                    num_parts=num_parts,
                )
            )
            multipart_upload = create_multipart_for_dataset_response.multipart_upload
        else:
            raise ValueError(
                "Invalid artifact type - both `artifact_version_id` and `dataset_fqn` both are None"
            )
        return multipart_upload

    def _multipart_upload(
        self,
        local_file: str,
        artifact_path: str,
        multipart_info: _FileMultiPartInfo,
        executor: ThreadPoolExecutor,
        progress_bar: Progress,
        abort_event: Optional[Event] = None,
    ):
        if progress_bar.disable:
            logger.info(
                "Uploading %s to %s using multipart upload",
                local_file,
                artifact_path,
            )

        multipart_upload = self._create_multipart_upload_for_identifier(
            artifact_identifier=self.artifact_identifier,
            path=artifact_path,
            num_parts=multipart_info.num_parts,
        )
        if (
            multipart_upload.storage_provider
            is MultiPartUploadStorageProvider.S3_COMPATIBLE
        ):
            s3_compatible_multipart_upload(
                multipart_upload=MultiPartUpload.parse_obj(multipart_upload.to_dict()),
                local_file=local_file,
                executor=executor,
                multipart_info=multipart_info,
                abort_event=abort_event,
                progress_bar=progress_bar,
                exception_class=MlFoundryException,  # type: ignore
            )
        elif (
            multipart_upload.storage_provider
            is MultiPartUploadStorageProvider.AZURE_BLOB
        ):
            azure_multi_part_upload(
                multipart_upload=MultiPartUpload.parse_obj(multipart_upload.to_dict()),
                local_file=local_file,
                executor=executor,
                multipart_info=multipart_info,
                abort_event=abort_event,
                progress_bar=progress_bar,
                exception_class=MlFoundryException,  # type: ignore
            )
        else:
            raise NotImplementedError()

    def _upload_file(
        self,
        local_file: str,
        artifact_path: str,
        multipart_info: _FileMultiPartInfo,
        progress_bar: Progress,
        signed_url: Optional[SignedURLDto] = None,
        abort_event: Optional[Event] = None,
        executor_for_multipart_upload: Optional[ThreadPoolExecutor] = None,
    ):
        if multipart_info.num_parts == 1:
            return self._normal_upload(
                local_file=local_file,
                artifact_path=artifact_path,
                signed_url=signed_url,
                abort_event=abort_event,
                progress_bar=progress_bar,
            )

        if not executor_for_multipart_upload:
            with ThreadPoolExecutor(
                max_workers=ENV_VARS.TFY_ARTIFACTS_UPLOAD_MAX_WORKERS
            ) as executor:
                return self._multipart_upload(
                    local_file=local_file,
                    artifact_path=artifact_path,
                    executor=executor,
                    multipart_info=multipart_info,
                    progress_bar=progress_bar,
                )

        return self._multipart_upload(
            local_file=local_file,
            artifact_path=artifact_path,
            executor=executor_for_multipart_upload,
            multipart_info=multipart_info,
            progress_bar=progress_bar,
        )

    def _upload(
        self,
        files_for_normal_upload: Sequence[Tuple[str, str, _FileMultiPartInfo]],
        files_for_multipart_upload: Sequence[Tuple[str, str, _FileMultiPartInfo]],
        progress: Optional[bool] = None,
    ):
        abort_event = Event()
        show_progress = _can_display_progress(progress)
        with Progress(
            "[progress.description]{task.description}",
            BarColumn(bar_width=None),
            "[progress.percentage]{task.percentage:>3.0f}%",
            DownloadColumn(),
            TransferSpeedColumn(),
            TimeRemainingColumn(),
            TimeElapsedColumn(),
            refresh_per_second=1,
            disable=not show_progress,
            expand=True,
        ) as progress_bar, ThreadPoolExecutor(
            max_workers=ENV_VARS.TFY_ARTIFACTS_UPLOAD_MAX_WORKERS
        ) as executor:
            futures: List[Future] = []
            # Note: While this batching is beneficial when there is a large number of files, there is also
            # a rare case risk of the signed url expiring before a request is made to it
            _batch_size = _GENERATE_SIGNED_URL_BATCH_SIZE
            for start_idx in range(0, len(files_for_normal_upload), _batch_size):
                end_idx = min(start_idx + _batch_size, len(files_for_normal_upload))
                if _any_future_has_failed(futures):
                    break
                logger.debug("Generating write signed urls for a batch ...")
                remote_file_paths = [
                    files_for_normal_upload[idx][0] for idx in range(start_idx, end_idx)
                ]
                signed_urls = self.get_signed_urls_for_write(
                    artifact_identifier=self.artifact_identifier,
                    paths=remote_file_paths,
                )
                for idx, signed_url in zip(range(start_idx, end_idx), signed_urls):
                    (
                        upload_path,
                        local_file,
                        multipart_info,
                    ) = files_for_normal_upload[idx]
                    future = executor.submit(
                        self._upload_file,
                        local_file=local_file,
                        artifact_path=upload_path,
                        multipart_info=multipart_info,
                        signed_url=signed_url,
                        abort_event=abort_event,
                        executor_for_multipart_upload=None,
                        progress_bar=progress_bar,
                    )
                    futures.append(future)

            done, not_done = wait(futures, return_when=FIRST_EXCEPTION)
            if len(not_done) > 0:
                abort_event.set()
            for future in not_done:
                future.cancel()
            for future in done:
                if future.exception() is not None:
                    raise future.exception()

            for (
                upload_path,
                local_file,
                multipart_info,
            ) in files_for_multipart_upload:
                self._upload_file(
                    local_file=local_file,
                    artifact_path=upload_path,
                    signed_url=None,
                    multipart_info=multipart_info,
                    executor_for_multipart_upload=executor,
                    progress_bar=progress_bar,
                )

    def _add_file_for_upload(
        self,
        local_file: str,
        artifact_path: str,
        files_for_normal_upload: List[Tuple[str, str, _FileMultiPartInfo]],
        files_for_multipart_upload: List[Tuple[str, str, _FileMultiPartInfo]],
    ):
        local_file = os.path.realpath(local_file)
        if os.path.isdir(local_file):
            raise MlFoundryException(
                "Cannot log a directory as an artifact. Use `log_artifacts` instead"
            )
        upload_path = artifact_path
        upload_path = upload_path.lstrip(posixpath.sep)
        multipart_info = decide_file_parts(local_file)
        if multipart_info.num_parts == 1:
            files_for_normal_upload.append((upload_path, local_file, multipart_info))
        else:
            files_for_multipart_upload.append((upload_path, local_file, multipart_info))

    def _add_dir_for_upload(
        self,
        local_dir: str,
        artifact_path: str,
        files_for_normal_upload: List[Tuple[str, str, _FileMultiPartInfo]],
        files_for_multipart_upload: List[Tuple[str, str, _FileMultiPartInfo]],
    ):
        local_dir = os.path.realpath(local_dir)
        if not os.path.isdir(local_dir):
            raise MlFoundryException(
                "Cannot log a file as a directory. Use `log_artifact` instead"
            )
        dest_path = artifact_path
        dest_path = dest_path.lstrip(posixpath.sep)

        for root, _, file_names in os.walk(local_dir):
            upload_path = dest_path
            if root != local_dir:
                rel_path = os.path.relpath(root, local_dir)
                rel_path = relative_path_to_artifact_path(rel_path)
                upload_path = posixpath.join(dest_path, rel_path)
            for file_name in file_names:
                local_file = os.path.join(root, file_name)
                self._add_file_for_upload(
                    local_file=local_file,
                    artifact_path=upload_path,
                    files_for_normal_upload=files_for_normal_upload,
                    files_for_multipart_upload=files_for_multipart_upload,
                )

    def log_artifacts(
        self,
        src_dest_pairs: Sequence[Tuple[str, Optional[str]]],
        progress: Optional[bool] = None,
    ):
        files_for_normal_upload: List[Tuple[str, str, _FileMultiPartInfo]] = []
        files_for_multipart_upload: List[Tuple[str, str, _FileMultiPartInfo]] = []
        for src, dest in src_dest_pairs:
            if os.path.isdir(src):
                self._add_dir_for_upload(
                    local_dir=src,
                    artifact_path=dest or "",
                    files_for_normal_upload=files_for_normal_upload,
                    files_for_multipart_upload=files_for_multipart_upload,
                )
            else:
                self._add_file_for_upload(
                    local_file=src,
                    artifact_path=dest or "",
                    files_for_normal_upload=files_for_normal_upload,
                    files_for_multipart_upload=files_for_multipart_upload,
                )
        self._upload(
            files_for_normal_upload=files_for_normal_upload,
            files_for_multipart_upload=files_for_multipart_upload,
            progress=progress,
        )

    def _list_files(
        self, artifact_identifier: ArtifactIdentifier, path, page_size, page_token
    ) -> Union[ListFilesForDatasetResponseDto, ListFilesForArtifactVersionsResponseDto]:
        if artifact_identifier.dataset_fqn:
            return self._mlfoundry_artifacts_api.list_files_for_dataset_post(
                list_files_for_dataset_request_dto=ListFilesForDatasetRequestDto(
                    dataset_fqn=artifact_identifier.dataset_fqn,
                    path=path,
                    max_results=page_size,
                    page_token=page_token,
                )
            )
        else:
            return self._mlfoundry_artifacts_api.list_files_for_artifact_version_post(
                list_files_for_artifact_version_request_dto=ListFilesForArtifactVersionRequestDto(
                    id=str(artifact_identifier.artifact_version_id),
                    path=path,
                    max_results=page_size,
                    page_token=page_token,
                )
            )

    def list_artifacts(
        self, path=None, page_size=_LIST_FILES_PAGE_SIZE, **kwargs
    ) -> Iterator[FileInfoDto]:
        page_token = None
        started = False
        while not started or page_token is not None:
            started = True
            page = self._list_files(
                artifact_identifier=self.artifact_identifier,
                path=path,
                page_size=page_size,
                page_token=page_token,
            )
            for file_info in page.files:
                yield file_info
            page_token = page.next_page_token

    def _is_directory(self, artifact_path):
        # TODO: Ideally server should return a flag to indicate if it is a directory
        # For now we can rely on list_artifacts because generally in cloud buckets
        # directories are not stored as separate entities - empty directories ideally do not exist
        for _ in self.list_artifacts(artifact_path, page_size=3):
            return True
        return False

    def _create_download_destination(
        self, src_artifact_path: str, dst_local_dir_path: str
    ) -> str:
        """
        Creates a local filesystem location to be used as a destination for downloading the artifact
        specified by `src_artifact_path`. The destination location is a subdirectory of the
        specified `dst_local_dir_path`, which is determined according to the structure of
        `src_artifact_path`. For example, if `src_artifact_path` is `dir1/file1.txt`, then the
        resulting destination path is `<dst_local_dir_path>/dir1/file1.txt`. Local directories are
        created for the resulting destination location if they do not exist.

        :param src_artifact_path: A relative, POSIX-style path referring to an artifact stored
                                  within the repository's artifact root location.
                                  `src_artifact_path` should be specified relative to the
                                  repository's artifact root location.
        :param dst_local_dir_path: The absolute path to a local filesystem directory in which the
                                   local destination path will be contained. The local destination
                                   path may be contained in a subdirectory of `dst_root_dir` if
                                   `src_artifact_path` contains subdirectories.
        :return: The absolute path to a local filesystem location to be used as a destination
                 for downloading the artifact specified by `src_artifact_path`.
        """
        src_artifact_path = src_artifact_path.rstrip(
            "/"
        )  # Ensure correct dirname for trailing '/'
        dirpath = posixpath.dirname(src_artifact_path)
        local_dir_path = os.path.join(dst_local_dir_path, dirpath)
        local_file_path = os.path.join(dst_local_dir_path, src_artifact_path)
        if not os.path.exists(local_dir_path):
            os.makedirs(local_dir_path, exist_ok=True)
        return local_file_path

    # noinspection PyMethodOverriding
    def _download_file(
        self,
        remote_file_path: str,
        local_path: str,
        progress_bar: Optional[Progress],
        signed_url: Optional[SignedURLDto],
        abort_event: Optional[Event] = None,
    ):
        if not remote_file_path:
            raise MlFoundryException(
                f"remote_file_path cannot be None or empty str {remote_file_path}"
            )
        if not signed_url:
            signed_url = self.get_signed_urls_for_read(
                artifact_identifier=self.artifact_identifier, paths=[remote_file_path]
            )[0]

        if progress_bar is None or not progress_bar.disable:
            logger.info("Downloading %s to %s", remote_file_path, local_path)

        if progress_bar is not None:
            download_progress_bar = progress_bar.add_task(
                f"[green]Downloading to {remote_file_path}:", start=True
            )

        def callback(chunk, total_file_size):
            if progress_bar is not None:
                progress_bar.update(
                    download_progress_bar,
                    advance=chunk,
                    total=total_file_size,
                )
            if abort_event and abort_event.is_set():
                raise Exception("aborting download")

        _download_file_using_http_uri(
            http_uri=signed_url.signed_url,
            download_path=local_path,
            callback=callback,
        )
        logger.debug("Downloaded %s to %s", remote_file_path, local_path)

    def _download_artifact(
        self,
        src_artifact_path: str,
        dst_local_dir_path: str,
        signed_url: Optional[SignedURLDto],
        progress_bar: Optional[Progress] = None,
        abort_event=None,
    ) -> str:
        """
        Download the file artifact specified by `src_artifact_path` to the local filesystem
        directory specified by `dst_local_dir_path`.
        :param src_artifact_path: A relative, POSIX-style path referring to a file artifact
                                    stored within the repository's artifact root location.
                                    `src_artifact_path` should be specified relative to the
                                    repository's artifact root location.
        :param dst_local_dir_path: Absolute path of the local filesystem destination directory
                                    to which to download the specified artifact. The downloaded
                                    artifact may be written to a subdirectory of
                                    `dst_local_dir_path` if `src_artifact_path` contains
                                    subdirectories.
        :param progress_bar: An instance of a Rich progress bar used to visually display the
                              progress of the file download.
        :return: A local filesystem path referring to the downloaded file.
        """
        local_destination_file_path = self._create_download_destination(
            src_artifact_path=src_artifact_path, dst_local_dir_path=dst_local_dir_path
        )
        self._download_file(
            remote_file_path=src_artifact_path,
            local_path=local_destination_file_path,
            signed_url=signed_url,
            abort_event=abort_event,
            progress_bar=progress_bar,
        )
        return local_destination_file_path

    def _get_file_paths_recur(self, src_artifact_dir_path, dst_local_dir_path):
        local_dir = os.path.join(dst_local_dir_path, src_artifact_dir_path)
        # prevent infinite loop, sometimes the dir is recursively included
        # TODO (chiragjn): Check why and when this happens
        dir_content = [
            file_info
            for file_info in self.list_artifacts(src_artifact_dir_path)
            if file_info.path != "." and file_info.path != src_artifact_dir_path
        ]
        if not dir_content:  # empty dir
            if not os.path.exists(local_dir):
                os.makedirs(local_dir, exist_ok=True)
        else:
            for file_info in dir_content:
                if file_info.is_dir:
                    yield from self._get_file_paths_recur(
                        src_artifact_dir_path=file_info.path,
                        dst_local_dir_path=dst_local_dir_path,
                    )
                else:
                    yield file_info.path, dst_local_dir_path

    def download_artifacts(  # noqa: C901
        self,
        artifact_path: str,
        dst_path: Optional[str] = None,
        overwrite: bool = False,
        progress: Optional[bool] = None,
    ) -> str:
        """
        Download an artifact file or directory to a local directory if applicable, and return a
        local path for it. The caller is responsible for managing the lifecycle of the downloaded artifacts.

        Args:
            artifact_path: Relative source path to the desired artifacts.
            dst_path: Absolute path of the local filesystem destination directory to which to
                download the specified artifacts. This directory must already exist.
                If unspecified, the artifacts will either be downloaded to a new
                uniquely-named directory.
            overwrite: if to overwrite the files at/inside `dst_path` if they exist
            progress: Show or hide progress bar

        Returns:
            str: Absolute path of the local filesystem location containing the desired artifacts.
        """

        show_progress = _can_display_progress(user_choice=progress)

        is_dir_temp = False
        if dst_path is None:
            dst_path = tempfile.mkdtemp()
            is_dir_temp = True

        dst_path = os.path.abspath(dst_path)
        if is_dir_temp:
            logger.info(
                f"Using temporary directory {dst_path} as the download directory"
            )

        if not os.path.exists(dst_path):
            raise MlFoundryException(
                message=(
                    "The destination path for downloaded artifacts does not"
                    " exist! Destination path: {dst_path}".format(dst_path=dst_path)
                ),
            )
        elif not os.path.isdir(dst_path):
            raise MlFoundryException(
                message=(
                    "The destination path for downloaded artifacts must be a directory!"
                    " Destination path: {dst_path}".format(dst_path=dst_path)
                ),
            )

        progress_bar = Progress(
            "[progress.description]{task.description}",
            BarColumn(),
            "[progress.percentage]{task.percentage:>3.0f}%",
            DownloadColumn(),
            TransferSpeedColumn(),
            TimeRemainingColumn(),
            TimeElapsedColumn(),
            refresh_per_second=1,
            disable=not show_progress,
            expand=True,
        )

        try:
            progress_bar.start()
            # empty string is treated as root directory
            if not artifact_path or self._is_directory(artifact_path):
                futures: List[Future] = []
                file_paths: List[Tuple[str, str]] = []
                abort_event = Event()

                # Check if any file is being overwritten before downloading them
                for file_path, download_dest_path in self._get_file_paths_recur(
                    src_artifact_dir_path=artifact_path, dst_local_dir_path=dst_path
                ):
                    final_file_path = os.path.join(download_dest_path, file_path)

                    # There would be no overwrite if temp directory is being used
                    if (
                        not is_dir_temp
                        and os.path.exists(final_file_path)
                        and not overwrite
                    ):
                        raise MlFoundryException(
                            f"File already exists at {final_file_path}, aborting download "
                            f"(set `overwrite` flag to overwrite this and any subsequent files)"
                        )
                    file_paths.append((file_path, download_dest_path))

                with ThreadPoolExecutor(
                    max_workers=ENV_VARS.TFY_ARTIFACTS_DOWNLOAD_MAX_WORKERS
                ) as executor:
                    # Note: While this batching is beneficial when there is a large number of files, there is also
                    # a rare case risk of the signed url expiring before a request is made to it
                    batch_size = _GENERATE_SIGNED_URL_BATCH_SIZE
                    for start_idx in range(0, len(file_paths), batch_size):
                        end_idx = min(start_idx + batch_size, len(file_paths))
                        if _any_future_has_failed(futures):
                            break
                        logger.debug("Generating read signed urls for a batch ...")
                        remote_file_paths = [
                            file_paths[idx][0] for idx in range(start_idx, end_idx)
                        ]
                        signed_urls = self.get_signed_urls_for_read(
                            artifact_identifier=self.artifact_identifier,
                            paths=remote_file_paths,
                        )
                        for idx, signed_url in zip(
                            range(start_idx, end_idx), signed_urls
                        ):
                            file_path, download_dest_path = file_paths[idx]
                            future = executor.submit(
                                self._download_artifact,
                                src_artifact_path=file_path,
                                dst_local_dir_path=download_dest_path,
                                signed_url=signed_url,
                                abort_event=abort_event,
                                progress_bar=progress_bar,
                            )
                            futures.append(future)

                    done, not_done = wait(futures, return_when=FIRST_EXCEPTION)
                    if len(not_done) > 0:
                        abort_event.set()
                    for future in not_done:
                        future.cancel()
                    for future in done:
                        if future.exception() is not None:
                            raise future.exception()

                    output_dir = os.path.join(dst_path, artifact_path)
                    return output_dir
            else:
                return self._download_artifact(
                    src_artifact_path=artifact_path,
                    dst_local_dir_path=dst_path,
                    signed_url=None,
                    progress_bar=progress_bar,
                )
        except Exception as err:
            if is_dir_temp:
                logger.info(
                    f"Error encountered, removing temporary download directory at {dst_path}"
                )
                rmtree(dst_path)  # remove temp directory alongside it's contents
            raise err

        finally:
            progress_bar.stop()



# coding: utf-8

"""
FastAPI

No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

The version of the OpenAPI document: 0.1.0
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations

import json
import pprint
import re  # noqa: F401

from truefoundry.ml.autogen.client.models.mime_type import MimeType
from truefoundry.pydantic_v1 import BaseModel, Field, StrictStr, constr, validator


class BlobStorageReference(BaseModel):
    """
    +usage=Blob Storage Location +label=Blob Storage Location +docs=Defines the structure for blob storage content references, including type, path, and mime_type  # noqa: E501
    """

    type: StrictStr = Field(...)
    path: constr(strict=True) = Field(
        default=...,
        description="+label=Path +usage=Path to the content in blob storage",
    )
    mime_type: MimeType = Field(
        default=..., description="+label=MIME Type +usage=MIME type of the content"
    )
    __properties = ["type", "path", "mime_type"]

    @validator("type")
    def type_validate_enum(cls, value):
        """Validates the enum"""
        if value not in ("blob-storage",):
            raise ValueError("must be one of enum values ('blob-storage')")
        return value

    @validator("path")
    def path_validate_regular_expression(cls, value):
        """Validates the regular expression"""
        if not re.match(r"^.{1,}$", value):
            raise ValueError(r"must validate the regular expression /^.{1,}$/")
        return value

    class Config:
        """Pydantic configuration"""

        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> BlobStorageReference:
        """Create an instance of BlobStorageReference from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True, exclude={}, exclude_none=True)
        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> BlobStorageReference:
        """Create an instance of BlobStorageReference from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return BlobStorageReference.parse_obj(obj)

        _obj = BlobStorageReference.parse_obj(
            {
                "type": obj.get("type"),
                "path": obj.get("path"),
                "mime_type": obj.get("mime_type"),
            }
        )
        return _obj



import functools

from truefoundry.common.constants import (
    SERVICEFOUNDRY_CLIENT_MAX_RETRIES,
    VERSION_PREFIX,
)
from truefoundry.common.exceptions import HttpRequestException
from truefoundry.common.request_utils import (
    http_request,
    request_handling,
    requests_retry_session,
)
from truefoundry.common.servicefoundry_client import (
    ServiceFoundryServiceClient as BaseServiceFoundryServiceClient,
)
from truefoundry.ml.exceptions import MlFoundryException


class ServiceFoundryServiceClient(BaseServiceFoundryServiceClient):
    def __init__(self, tfy_host: str, token: str):
        super().__init__(tfy_host=tfy_host)
        self._token = token

    @functools.cached_property
    def _min_cli_version_required(self) -> str:
        # TODO (chiragjn): read the mlfoundry min cli version from the config?
        return self.python_sdk_config.truefoundry_cli_min_version

    def get_integration_from_id(self, integration_id: str):
        integration_id = integration_id or ""
        session = requests_retry_session(retries=SERVICEFOUNDRY_CLIENT_MAX_RETRIES)
        response = http_request(
            method="get",
            url=f"{self._api_server_url}/{VERSION_PREFIX}/provider-accounts/provider-integrations",
            token=self._token,
            timeout=3,
            params={"id": integration_id, "type": "blob-storage"},
            session=session,
        )

        try:
            data = request_handling(response)
            assert isinstance(data, dict)
        except HttpRequestException as he:
            raise MlFoundryException(
                f"Failed to get storage integration from id: {integration_id}. Error: {he.message}",
                status_code=he.status_code,
            ) from None
        except Exception as e:
            raise MlFoundryException(
                f"Failed to get storage integration from id: {integration_id}. Error: {str(e)}"
            ) from None

        # TODO (chiragjn): Parse this using Pydantic
        if (
            data.get("providerIntegrations")
            and len(data["providerIntegrations"]) > 0
            and data["providerIntegrations"][0]
        ):
            return data["providerIntegrations"][0]
        else:
            raise MlFoundryException(
                f"Invalid storage integration id: {integration_id}"
            )



# coding: utf-8

"""
FastAPI

No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

The version of the OpenAPI document: 0.1.0
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations

import json
import pprint
import re  # noqa: F401
from typing import Dict, Optional

from truefoundry.ml.autogen.client.models.chat_prompt_messages_inner import (
    ChatPromptMessagesInner,
)
from truefoundry.ml.autogen.client.models.model_configuration import ModelConfiguration
from truefoundry.pydantic_v1 import (
    BaseModel,
    Field,
    StrictStr,
    conlist,
    constr,
    validator,
)


class ChatPrompt(BaseModel):
    """
    Chat Prompt artifact.  # noqa: E501
    """

    name: constr(strict=True) = Field(
        default=...,
        description="+label=Prompt Name +icon=fa-desktop:#326ce5 +message=Alphanumeric word, may contain '-' with a maximum length of 256 characters +sort=1000 +usage=Name of the Prompt. This uniquely identifies it in the workspace. +docs=The unique name for the prompt, consisting of alphanumeric characters and dashes, max length 100.",
    )
    description: Optional[constr(strict=True, max_length=512, min_length=1)] = Field(
        default=None,
        description="+label=Description +icon=fa-desktop:#326ce5 +message=Description in a maximum of 512 characters +sort=2000 +usage=Description of the prompt +docs=An optional description for the prompt, with a maximum length of 512 characters. +optional",
    )
    model_configuration: ModelConfiguration = Field(...)
    metadata: Dict[str, StrictStr] = Field(
        default=...,
        description="+label=Metadata +icon=fa-info-circle:#326ce5 +message=Additional metadata for the prompt +sort=7000 +usage=Define additional metadata for the prompt +docs=Key-value pairs to store additional metadata related to the prompt.",
    )
    type: Optional[StrictStr] = "chat_prompt"
    messages: conlist(ChatPromptMessagesInner) = Field(
        default=...,
        description="+sort=4000 +usage=Chat completion messages +label=Messages in the chat conversation +message=Chat completion messages +usage=List of messages in the chat conversation, must be non-empty +docs=Messages that define the chat conversation, including system, assistant, and user messages.",
    )
    variables: Optional[Dict[str, StrictStr]] = Field(
        default=None,
        description="+label=Variables +usage=Variables for the chat completion messages to be used in the prompt messages +sort=5000",
    )
    __properties = [
        "name",
        "description",
        "model_configuration",
        "metadata",
        "type",
        "messages",
        "variables",
    ]

    @validator("name")
    def name_validate_regular_expression(cls, value):
        """Validates the regular expression"""
        if not re.match(r"^[a-zA-Z][a-zA-Z0-9-_]{0,254}[a-zA-Z0-9]$", value):
            raise ValueError(
                r"must validate the regular expression /^[a-zA-Z][a-zA-Z0-9-_]{0,254}[a-zA-Z0-9]$/"
            )
        return value

    @validator("type")
    def type_validate_enum(cls, value):
        """Validates the enum"""
        if value is None:
            return value

        if value not in ("chat_prompt",):
            raise ValueError("must be one of enum values ('chat_prompt')")
        return value

    class Config:
        """Pydantic configuration"""

        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> ChatPrompt:
        """Create an instance of ChatPrompt from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True, exclude={}, exclude_none=True)
        # override the default output from truefoundry.pydantic_v1 by calling `to_dict()` of model_configuration
        if self.model_configuration:
            _dict["model_configuration"] = self.model_configuration.to_dict()
        # override the default output from truefoundry.pydantic_v1 by calling `to_dict()` of each item in messages (list)
        _items = []
        if self.messages:
            for _item in self.messages:
                if _item:
                    _items.append(_item.to_dict())
            _dict["messages"] = _items
        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> ChatPrompt:
        """Create an instance of ChatPrompt from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return ChatPrompt.parse_obj(obj)

        _obj = ChatPrompt.parse_obj(
            {
                "name": obj.get("name"),
                "description": obj.get("description"),
                "model_configuration": ModelConfiguration.from_dict(
                    obj.get("model_configuration")
                )
                if obj.get("model_configuration") is not None
                else None,
                "metadata": obj.get("metadata"),
                "type": obj.get("type")
                if obj.get("type") is not None
                else "chat_prompt",
                "messages": [
                    ChatPromptMessagesInner.from_dict(_item)
                    for _item in obj.get("messages")
                ]
                if obj.get("messages") is not None
                else None,
                "variables": obj.get("variables"),
            }
        )
        return _obj



from truefoundry.workflow import PythonTaskConfig, TaskPythonBuild, task, workflow


@task(
    task_config=PythonTaskConfig(
        image=TaskPythonBuild(
            python_version="3.9",
            pip_packages=["truefoundry[workflow]"],
        ),
        service_account="tfy-workflows-sa",
    )
)
def say_hello() -> str:
    return "Hello, World!"


@workflow
def hello_world_wf() -> str:
    res = say_hello()
    return res



import re
from typing import Union


def get_application_fqn_from_deployment_fqn(deployment_fqn: str) -> str:
    if not re.search(r":\d+$", deployment_fqn):
        raise ValueError(
            "Invalid `deployment_fqn` format. A deployment fqn is supposed to end with a version number"
        )
    application_fqn, _ = deployment_fqn.rsplit(":", 1)
    return application_fqn


def get_deployment_fqn_from_application_fqn(
    application_fqn: str, version: Union[str, int]
) -> str:
    return f"{application_fqn}:{version}"


def find_list_paths(data, parent_key="", sep="."):
    list_paths = []
    if isinstance(data, dict):
        for key, value in data.items():
            new_key = f"{parent_key}{sep}{key}" if parent_key else key
            list_paths.extend(find_list_paths(value, new_key, sep))
    elif isinstance(data, list):
        list_paths.append(parent_key)
        for i, value in enumerate(data):
            new_key = f"{parent_key}[{i}]"
            list_paths.extend(find_list_paths(value, new_key, sep))
    return list_paths



# coding: utf-8

# flake8: noqa

"""
FastAPI

No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

The version of the OpenAPI document: 0.1.0
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

__version__ = "0.1.0"

# import apis into sdk package
from truefoundry.ml.autogen.client.api.auth_api import AuthApi
from truefoundry.ml.autogen.client.api.deprecated_api import DeprecatedApi
from truefoundry.ml.autogen.client.api.experiments_api import ExperimentsApi
from truefoundry.ml.autogen.client.api.generate_code_snippet_api import (
    GenerateCodeSnippetApi,
)
from truefoundry.ml.autogen.client.api.health_api import HealthApi
from truefoundry.ml.autogen.client.api.metrics_api import MetricsApi
from truefoundry.ml.autogen.client.api.mlfoundry_artifacts_api import (
    MlfoundryArtifactsApi,
)
from truefoundry.ml.autogen.client.api.run_artifacts_api import RunArtifactsApi
from truefoundry.ml.autogen.client.api.runs_api import RunsApi

# import ApiClient
from truefoundry.ml.autogen.client.api_response import ApiResponse
from truefoundry.ml.autogen.client.api_client import ApiClient
from truefoundry.ml.autogen.client.configuration import Configuration
from truefoundry.ml.autogen.client.exceptions import OpenApiException
from truefoundry.ml.autogen.client.exceptions import ApiTypeError
from truefoundry.ml.autogen.client.exceptions import ApiValueError
from truefoundry.ml.autogen.client.exceptions import ApiKeyError
from truefoundry.ml.autogen.client.exceptions import ApiAttributeError
from truefoundry.ml.autogen.client.exceptions import ApiException

# import models into sdk package
from truefoundry.ml.autogen.client.models.add_custom_metrics_to_model_version_request_dto import (
    AddCustomMetricsToModelVersionRequestDto,
)
from truefoundry.ml.autogen.client.models.add_features_to_model_version_request_dto import (
    AddFeaturesToModelVersionRequestDto,
)
from truefoundry.ml.autogen.client.models.agent import Agent
from truefoundry.ml.autogen.client.models.agent_app import AgentApp
from truefoundry.ml.autogen.client.models.agent_open_api_tool import AgentOpenAPITool
from truefoundry.ml.autogen.client.models.agent_open_api_tool_with_fqn import (
    AgentOpenAPIToolWithFQN,
)
from truefoundry.ml.autogen.client.models.agent_with_fqn import AgentWithFQN
from truefoundry.ml.autogen.client.models.artifact_dto import ArtifactDto
from truefoundry.ml.autogen.client.models.artifact_response_dto import (
    ArtifactResponseDto,
)
from truefoundry.ml.autogen.client.models.artifact_type import ArtifactType
from truefoundry.ml.autogen.client.models.artifact_version_dto import ArtifactVersionDto
from truefoundry.ml.autogen.client.models.artifact_version_manifest import (
    ArtifactVersionManifest,
)
from truefoundry.ml.autogen.client.models.artifact_version_response_dto import (
    ArtifactVersionResponseDto,
)
from truefoundry.ml.autogen.client.models.artifact_version_serialization_format import (
    ArtifactVersionSerializationFormat,
)
from truefoundry.ml.autogen.client.models.artifact_version_status import (
    ArtifactVersionStatus,
)
from truefoundry.ml.autogen.client.models.assistant_message import AssistantMessage
from truefoundry.ml.autogen.client.models.authorize_user_for_model_request_dto import (
    AuthorizeUserForModelRequestDto,
)
from truefoundry.ml.autogen.client.models.authorize_user_for_model_version_request_dto import (
    AuthorizeUserForModelVersionRequestDto,
)
from truefoundry.ml.autogen.client.models.blob_storage_reference import (
    BlobStorageReference,
)
from truefoundry.ml.autogen.client.models.body_get_search_runs_get import (
    BodyGetSearchRunsGet,
)
from truefoundry.ml.autogen.client.models.chat_prompt import ChatPrompt
from truefoundry.ml.autogen.client.models.chat_prompt_messages_inner import (
    ChatPromptMessagesInner,
)
from truefoundry.ml.autogen.client.models.columns_dto import ColumnsDto
from truefoundry.ml.autogen.client.models.command import Command
from truefoundry.ml.autogen.client.models.content import Content
from truefoundry.ml.autogen.client.models.content1 import Content1
from truefoundry.ml.autogen.client.models.content2 import Content2
from truefoundry.ml.autogen.client.models.content2_any_of_inner import (
    Content2AnyOfInner,
)
from truefoundry.ml.autogen.client.models.create_artifact_request_dto import (
    CreateArtifactRequestDto,
)
from truefoundry.ml.autogen.client.models.create_artifact_response_dto import (
    CreateArtifactResponseDto,
)
from truefoundry.ml.autogen.client.models.create_artifact_version_request_dto import (
    CreateArtifactVersionRequestDto,
)
from truefoundry.ml.autogen.client.models.create_artifact_version_response_dto import (
    CreateArtifactVersionResponseDto,
)
from truefoundry.ml.autogen.client.models.create_dataset_request_dto import (
    CreateDatasetRequestDto,
)
from truefoundry.ml.autogen.client.models.create_experiment_request_dto import (
    CreateExperimentRequestDto,
)
from truefoundry.ml.autogen.client.models.create_experiment_response_dto import (
    CreateExperimentResponseDto,
)
from truefoundry.ml.autogen.client.models.create_model_version_request_dto import (
    CreateModelVersionRequestDto,
)
from truefoundry.ml.autogen.client.models.create_multi_part_upload_for_dataset_request_dto import (
    CreateMultiPartUploadForDatasetRequestDto,
)
from truefoundry.ml.autogen.client.models.create_multi_part_upload_for_dataset_response_dto import (
    CreateMultiPartUploadForDatasetResponseDto,
)
from truefoundry.ml.autogen.client.models.create_multi_part_upload_request_dto import (
    CreateMultiPartUploadRequestDto,
)
from truefoundry.ml.autogen.client.models.create_python_deployment_config_request_dto import (
    CreatePythonDeploymentConfigRequestDto,
)
from truefoundry.ml.autogen.client.models.create_python_deployment_config_response_dto import (
    CreatePythonDeploymentConfigResponseDto,
)
from truefoundry.ml.autogen.client.models.create_run_request_dto import (
    CreateRunRequestDto,
)
from truefoundry.ml.autogen.client.models.create_run_response_dto import (
    CreateRunResponseDto,
)
from truefoundry.ml.autogen.client.models.create_workflow_task_config_request_dto import (
    CreateWorkflowTaskConfigRequestDto,
)
from truefoundry.ml.autogen.client.models.dataset_dto import DatasetDto
from truefoundry.ml.autogen.client.models.dataset_response_dto import DatasetResponseDto
from truefoundry.ml.autogen.client.models.delete_artifact_versions_request_dto import (
    DeleteArtifactVersionsRequestDto,
)
from truefoundry.ml.autogen.client.models.delete_dataset_request_dto import (
    DeleteDatasetRequestDto,
)
from truefoundry.ml.autogen.client.models.delete_files_for_dataset_request_dto import (
    DeleteFilesForDatasetRequestDto,
)
from truefoundry.ml.autogen.client.models.delete_model_version_request_dto import (
    DeleteModelVersionRequestDto,
)
from truefoundry.ml.autogen.client.models.delete_run_request import DeleteRunRequest
from truefoundry.ml.autogen.client.models.delete_tag_request_dto import (
    DeleteTagRequestDto,
)
from truefoundry.ml.autogen.client.models.experiment_dto import ExperimentDto
from truefoundry.ml.autogen.client.models.experiment_id_request_dto import (
    ExperimentIdRequestDto,
)
from truefoundry.ml.autogen.client.models.experiment_response_dto import (
    ExperimentResponseDto,
)
from truefoundry.ml.autogen.client.models.experiment_tag_dto import ExperimentTagDto
from truefoundry.ml.autogen.client.models.export_deployment_files_request_dto import (
    ExportDeploymentFilesRequestDto,
)
from truefoundry.ml.autogen.client.models.external_artifact_source import (
    ExternalArtifactSource,
)
from truefoundry.ml.autogen.client.models.fast_ai_framework import FastAIFramework
from truefoundry.ml.autogen.client.models.file_info_dto import FileInfoDto
from truefoundry.ml.autogen.client.models.finalize_artifact_version_request_dto import (
    FinalizeArtifactVersionRequestDto,
)
from truefoundry.ml.autogen.client.models.framework import Framework
from truefoundry.ml.autogen.client.models.get_artifact_version_aliases_response_dto import (
    GetArtifactVersionAliasesResponseDto,
)
from truefoundry.ml.autogen.client.models.get_experiment_response_dto import (
    GetExperimentResponseDto,
)
from truefoundry.ml.autogen.client.models.get_latest_run_log_response_dto import (
    GetLatestRunLogResponseDto,
)
from truefoundry.ml.autogen.client.models.get_metric_history_response import (
    GetMetricHistoryResponse,
)
from truefoundry.ml.autogen.client.models.get_signed_url_for_dataset_write_request_dto import (
    GetSignedURLForDatasetWriteRequestDto,
)
from truefoundry.ml.autogen.client.models.get_signed_urls_for_artifact_version_read_request_dto import (
    GetSignedURLsForArtifactVersionReadRequestDto,
)
from truefoundry.ml.autogen.client.models.get_signed_urls_for_artifact_version_read_response_dto import (
    GetSignedURLsForArtifactVersionReadResponseDto,
)
from truefoundry.ml.autogen.client.models.get_signed_urls_for_artifact_version_write_request_dto import (
    GetSignedURLsForArtifactVersionWriteRequestDto,
)
from truefoundry.ml.autogen.client.models.get_signed_urls_for_artifact_version_write_response_dto import (
    GetSignedURLsForArtifactVersionWriteResponseDto,
)
from truefoundry.ml.autogen.client.models.get_signed_urls_for_dataset_read_request_dto import (
    GetSignedURLsForDatasetReadRequestDto,
)
from truefoundry.ml.autogen.client.models.get_signed_urls_for_dataset_read_response_dto import (
    GetSignedURLsForDatasetReadResponseDto,
)
from truefoundry.ml.autogen.client.models.get_signed_urls_for_dataset_write_response_dto import (
    GetSignedURLsForDatasetWriteResponseDto,
)
from truefoundry.ml.autogen.client.models.get_tenant_id_response_dto import (
    GetTenantIdResponseDto,
)
from truefoundry.ml.autogen.client.models.gluon_framework import GluonFramework
from truefoundry.ml.autogen.client.models.h2_o_framework import H2OFramework
from truefoundry.ml.autogen.client.models.http_validation_error import (
    HTTPValidationError,
)
from truefoundry.ml.autogen.client.models.image_content_part import ImageContentPart
from truefoundry.ml.autogen.client.models.image_url import ImageUrl
from truefoundry.ml.autogen.client.models.infer_method_name import InferMethodName
from truefoundry.ml.autogen.client.models.internal_metadata import InternalMetadata
from truefoundry.ml.autogen.client.models.keras_framework import KerasFramework
from truefoundry.ml.autogen.client.models.latest_run_log_dto import LatestRunLogDto
from truefoundry.ml.autogen.client.models.library_name import LibraryName
from truefoundry.ml.autogen.client.models.light_gbm_framework import LightGBMFramework
from truefoundry.ml.autogen.client.models.list_artifact_versions_request_dto import (
    ListArtifactVersionsRequestDto,
)
from truefoundry.ml.autogen.client.models.list_artifact_versions_response_dto import (
    ListArtifactVersionsResponseDto,
)
from truefoundry.ml.autogen.client.models.list_artifacts_request_dto import (
    ListArtifactsRequestDto,
)
from truefoundry.ml.autogen.client.models.list_artifacts_response_dto import (
    ListArtifactsResponseDto,
)
from truefoundry.ml.autogen.client.models.list_colums_response_dto import (
    ListColumsResponseDto,
)
from truefoundry.ml.autogen.client.models.list_datasets_request_dto import (
    ListDatasetsRequestDto,
)
from truefoundry.ml.autogen.client.models.list_datasets_response_dto import (
    ListDatasetsResponseDto,
)
from truefoundry.ml.autogen.client.models.list_experiments_response_dto import (
    ListExperimentsResponseDto,
)
from truefoundry.ml.autogen.client.models.list_files_for_artifact_version_request_dto import (
    ListFilesForArtifactVersionRequestDto,
)
from truefoundry.ml.autogen.client.models.list_files_for_artifact_versions_response_dto import (
    ListFilesForArtifactVersionsResponseDto,
)
from truefoundry.ml.autogen.client.models.list_files_for_dataset_request_dto import (
    ListFilesForDatasetRequestDto,
)
from truefoundry.ml.autogen.client.models.list_files_for_dataset_response_dto import (
    ListFilesForDatasetResponseDto,
)
from truefoundry.ml.autogen.client.models.list_latest_run_logs_response_dto import (
    ListLatestRunLogsResponseDto,
)
from truefoundry.ml.autogen.client.models.list_metric_history_request_dto import (
    ListMetricHistoryRequestDto,
)
from truefoundry.ml.autogen.client.models.list_metric_history_response_dto import (
    ListMetricHistoryResponseDto,
)
from truefoundry.ml.autogen.client.models.list_model_version_response_dto import (
    ListModelVersionResponseDto,
)
from truefoundry.ml.autogen.client.models.list_model_versions_request_dto import (
    ListModelVersionsRequestDto,
)
from truefoundry.ml.autogen.client.models.list_models_request_dto import (
    ListModelsRequestDto,
)
from truefoundry.ml.autogen.client.models.list_models_response_dto import (
    ListModelsResponseDto,
)
from truefoundry.ml.autogen.client.models.list_run_artifacts_response_dto import (
    ListRunArtifactsResponseDto,
)
from truefoundry.ml.autogen.client.models.list_run_logs_response_dto import (
    ListRunLogsResponseDto,
)
from truefoundry.ml.autogen.client.models.log_batch_request_dto import (
    LogBatchRequestDto,
)
from truefoundry.ml.autogen.client.models.log_metric_request_dto import (
    LogMetricRequestDto,
)
from truefoundry.ml.autogen.client.models.log_param_request_dto import (
    LogParamRequestDto,
)
from truefoundry.ml.autogen.client.models.manifest import Manifest
from truefoundry.ml.autogen.client.models.method import Method
from truefoundry.ml.autogen.client.models.metric_collection_dto import (
    MetricCollectionDto,
)
from truefoundry.ml.autogen.client.models.metric_dto import MetricDto
from truefoundry.ml.autogen.client.models.mime_type import MimeType
from truefoundry.ml.autogen.client.models.model_configuration import ModelConfiguration
from truefoundry.ml.autogen.client.models.model_dto import ModelDto
from truefoundry.ml.autogen.client.models.model_response_dto import ModelResponseDto
from truefoundry.ml.autogen.client.models.model_server import ModelServer
from truefoundry.ml.autogen.client.models.model_version_dto import ModelVersionDto
from truefoundry.ml.autogen.client.models.model_version_environment import (
    ModelVersionEnvironment,
)
from truefoundry.ml.autogen.client.models.model_version_manifest import (
    ModelVersionManifest,
)
from truefoundry.ml.autogen.client.models.model_version_response_dto import (
    ModelVersionResponseDto,
)
from truefoundry.ml.autogen.client.models.multi_part_upload_dto import (
    MultiPartUploadDto,
)
from truefoundry.ml.autogen.client.models.multi_part_upload_response_dto import (
    MultiPartUploadResponseDto,
)
from truefoundry.ml.autogen.client.models.multi_part_upload_storage_provider import (
    MultiPartUploadStorageProvider,
)
from truefoundry.ml.autogen.client.models.notify_artifact_version_failure_dto import (
    NotifyArtifactVersionFailureDto,
)
from truefoundry.ml.autogen.client.models.onnx_framework import ONNXFramework
from truefoundry.ml.autogen.client.models.openapi_spec import OpenapiSpec
from truefoundry.ml.autogen.client.models.paddle_framework import PaddleFramework
from truefoundry.ml.autogen.client.models.param_dto import ParamDto
from truefoundry.ml.autogen.client.models.parameters import Parameters
from truefoundry.ml.autogen.client.models.py_torch_framework import PyTorchFramework
from truefoundry.ml.autogen.client.models.resolve_agent_app_response_dto import (
    ResolveAgentAppResponseDto,
)
from truefoundry.ml.autogen.client.models.restore_run_request_dto import (
    RestoreRunRequestDto,
)
from truefoundry.ml.autogen.client.models.run_data_dto import RunDataDto
from truefoundry.ml.autogen.client.models.run_dto import RunDto
from truefoundry.ml.autogen.client.models.run_info_dto import RunInfoDto
from truefoundry.ml.autogen.client.models.run_log_dto import RunLogDto
from truefoundry.ml.autogen.client.models.run_log_input_dto import RunLogInputDto
from truefoundry.ml.autogen.client.models.run_response_dto import RunResponseDto
from truefoundry.ml.autogen.client.models.run_tag_dto import RunTagDto
from truefoundry.ml.autogen.client.models.search_runs_request_dto import (
    SearchRunsRequestDto,
)
from truefoundry.ml.autogen.client.models.search_runs_response_dto import (
    SearchRunsResponseDto,
)
from truefoundry.ml.autogen.client.models.set_experiment_tag_request_dto import (
    SetExperimentTagRequestDto,
)
from truefoundry.ml.autogen.client.models.set_tag_request_dto import SetTagRequestDto
from truefoundry.ml.autogen.client.models.signed_url_dto import SignedURLDto
from truefoundry.ml.autogen.client.models.sklearn_framework import SklearnFramework
from truefoundry.ml.autogen.client.models.sklearn_model_schema import SklearnModelSchema
from truefoundry.ml.autogen.client.models.sklearn_serialization_format import (
    SklearnSerializationFormat,
)
from truefoundry.ml.autogen.client.models.source import Source
from truefoundry.ml.autogen.client.models.source1 import Source1
from truefoundry.ml.autogen.client.models.spa_cy_framework import SpaCyFramework
from truefoundry.ml.autogen.client.models.stats_models_framework import (
    StatsModelsFramework,
)
from truefoundry.ml.autogen.client.models.stop import Stop
from truefoundry.ml.autogen.client.models.store_run_logs_request_dto import (
    StoreRunLogsRequestDto,
)
from truefoundry.ml.autogen.client.models.subject import Subject
from truefoundry.ml.autogen.client.models.subject_type import SubjectType
from truefoundry.ml.autogen.client.models.system_message import SystemMessage
from truefoundry.ml.autogen.client.models.tensor_flow_framework import (
    TensorFlowFramework,
)
from truefoundry.ml.autogen.client.models.text import Text
from truefoundry.ml.autogen.client.models.text_content_part import TextContentPart
from truefoundry.ml.autogen.client.models.transformers_framework import (
    TransformersFramework,
)
from truefoundry.ml.autogen.client.models.trigger_job_run_config_request_dto import (
    TriggerJobRunConfigRequestDto,
)
from truefoundry.ml.autogen.client.models.trigger_job_run_config_response_dto import (
    TriggerJobRunConfigResponseDto,
)
from truefoundry.ml.autogen.client.models.true_foundry_artifact_source import (
    TrueFoundryArtifactSource,
)
from truefoundry.ml.autogen.client.models.update_artifact_version_request_dto import (
    UpdateArtifactVersionRequestDto,
)
from truefoundry.ml.autogen.client.models.update_dataset_request_dto import (
    UpdateDatasetRequestDto,
)
from truefoundry.ml.autogen.client.models.update_experiment_request_dto import (
    UpdateExperimentRequestDto,
)
from truefoundry.ml.autogen.client.models.update_model_version_request_dto import (
    UpdateModelVersionRequestDto,
)
from truefoundry.ml.autogen.client.models.update_run_request_dto import (
    UpdateRunRequestDto,
)
from truefoundry.ml.autogen.client.models.update_run_response_dto import (
    UpdateRunResponseDto,
)
from truefoundry.ml.autogen.client.models.url import Url
from truefoundry.ml.autogen.client.models.user_message import UserMessage
from truefoundry.ml.autogen.client.models.validate_external_storage_root_request_dto import (
    ValidateExternalStorageRootRequestDto,
)
from truefoundry.ml.autogen.client.models.validate_external_storage_root_response_dto import (
    ValidateExternalStorageRootResponseDto,
)
from truefoundry.ml.autogen.client.models.validation_error import ValidationError
from truefoundry.ml.autogen.client.models.validation_error_loc_inner import (
    ValidationErrorLocInner,
)
from truefoundry.ml.autogen.client.models.xg_boost_framework import XGBoostFramework
from truefoundry.ml.autogen.client.models.xg_boost_model_schema import (
    XGBoostModelSchema,
)
from truefoundry.ml.autogen.client.models.xg_boost_serialization_format import (
    XGBoostSerializationFormat,
)



from typing import List, Optional

from truefoundry.deploy.builder.constants import (
    BUILDKIT_SECRET_MOUNT_PIP_CONF_ID,
    BUILDKIT_SECRET_MOUNT_UV_CONF_ID,
)


def _get_id_from_buildkit_secret_value(value: str) -> Optional[str]:
    parts = value.split(",")
    secret_config = {}
    for part in parts:
        kv = part.split("=", 1)
        if len(kv) != 2:
            continue
        key, value = kv
        secret_config[key] = value

    if "id" in secret_config and "src" in secret_config:
        return secret_config["id"]

    return None


def has_python_package_manager_conf_secret(docker_build_extra_args: List[str]) -> bool:
    args = [arg.strip() for arg in docker_build_extra_args]
    for i, arg in enumerate(docker_build_extra_args):
        if (
            arg == "--secret"
            and i + 1 < len(args)
            and (
                _get_id_from_buildkit_secret_value(args[i + 1])
                in (BUILDKIT_SECRET_MOUNT_PIP_CONF_ID, BUILDKIT_SECRET_MOUNT_UV_CONF_ID)
            )
        ):
            return True
    return False



import os
from functools import partial
from pathlib import Path
from typing import List, Optional, Tuple

from truefoundry.deploy import Image, NvidiaGPU, Resources
from truefoundry.workflow import (
    ContainerTask,
    ContainerTaskConfig,
    FlyteDirectory,
    PythonTaskConfig,
    TaskPythonBuild,
    conditional,
    map_task,
    task,
    workflow,
)

cpu_task_config = PythonTaskConfig(
    image=TaskPythonBuild(
        python_version="3.9",
        pip_packages=["truefoundry[workflow]==0.4.1"],
    ),
    resources=Resources(cpu_request=0.45),
    service_account="tfy-workflows-sa",
)


# figure out naming
@task(task_config=cpu_task_config)
def should_train_tokenizer(tokenizer: str) -> bool:
    print("Should train tokenizer")
    return not bool(tokenizer)


@task(
    task_config=cpu_task_config,
    # currently we will not support the caching, but keeping here for now
    # cache=True,
    # cache_version="1.0",
)
def train_tokenizer() -> str:
    print("Training tokenizer")
    return "trained_tokenizer"


@task(
    task_config=PythonTaskConfig(
        image=TaskPythonBuild(
            python_version="3.9",
            pip_packages=["truefoundry[workflow]==0.4.1", "pynvml==11.5.0"],
            cuda_version="11.5-cudnn8",
        ),
        env={
            "NVIDIA_DRIVER_CAPABILITIES": "compute,utility",
            "NVIDIA_VISIBLE_DEVICES": "all",
        },
        resources=Resources(cpu_request=0.45, devices=[NvidiaGPU(name="T4", count=1)]),
        service_account="tfy-workflows-sa",
    ),
)
def train_model(tokenizer: str) -> Tuple[FlyteDirectory, str]:
    print("Training model")
    import flytekit
    # from pynvml import nvmlDeviceGetCount, nvmlInit

    # nvmlInit()
    # assert nvmlDeviceGetCount() > 0

    working_dir = flytekit.current_context().working_directory
    local_dir = Path(os.path.join(working_dir, "csv_files"))
    local_dir.mkdir(exist_ok=True)

    with open(os.path.join(local_dir, "model"), "w", encoding="utf-8") as f:
        f.write(tokenizer)

    return FlyteDirectory(path=str(local_dir)), "hello"


@task(task_config=cpu_task_config)
def get_validation_data() -> List[str]:
    print("Getting validation data")
    return ["foo", "bar", "baz"]


@task(task_config=cpu_task_config)
def validate_model(model: FlyteDirectory, tokenizer: str, validation_data: str) -> bool:
    print(validation_data)
    model_path = os.path.join(model, "model")
    with open(model_path, "r", encoding="utf-8") as f:
        return f.read() == tokenizer


@task(task_config=cpu_task_config)
def all_good(validations: List[bool]) -> bool:
    print("Validations", validations)
    return all(validations)


echo = ContainerTask(
    name="echo",
    task_config=ContainerTaskConfig(
        image=Image(
            image_uri="docker.io/bash:4.1",
            command=["echo", "hello"],
        ),
        service_account="tfy-workflows-sa",
    ),
)


@task(task_config=cpu_task_config)
def random(tokenizer: str) -> Tuple[FlyteDirectory, str]:
    print(tokenizer)
    return FlyteDirectory(path=""), "random"


@workflow(
    # execution_configs=[
    #     ExecutionConfig(
    #         schedule="*/10 * * * *",
    #     )
    # ]
)
def train(
    tokenizer: str = "",
    test_v2: str = "",
    optional_var: Optional[str] = "",
    default_v: Optional[str] = "hello1",
    default_v2: str = "hello2",
) -> bool:
    stt = should_train_tokenizer(tokenizer=tokenizer)
    model, t = (
        conditional("train_tokenizer")
        .if_(stt.is_true())
        .then(train_model(tokenizer=tokenizer))
        .else_()
        .then(random(tokenizer=tokenizer))
    )
    validation_task = partial(validate_model, model=model, tokenizer=t)
    validation_data = get_validation_data()
    validations = map_task(
        validation_task,
        concurrency=2,
    )(validation_data=validation_data)
    echo()
    return all_good(validations=validations)


if __name__ == "__main__":
    train()



# coding: utf-8

"""
FastAPI

No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

The version of the OpenAPI document: 0.1.0
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

import json
import re  # noqa: F401

from aenum import Enum


class MultiPartUploadStorageProvider(str, Enum):
    """
    An enumeration.
    """

    """
    allowed enum values
    """
    S3_COMPATIBLE = "S3_COMPATIBLE"
    AZURE_BLOB = "AZURE_BLOB"

    @classmethod
    def from_json(cls, json_str: str) -> MultiPartUploadStorageProvider:
        """Create an instance of MultiPartUploadStorageProvider from a JSON string"""
        return MultiPartUploadStorageProvider(json.loads(json_str))



# coding: utf-8

"""
FastAPI

No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

The version of the OpenAPI document: 0.1.0
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations

import json
import pprint
import re  # noqa: F401
from typing import Optional

from truefoundry.pydantic_v1 import (
    BaseModel,
    Field,
    StrictStr,
    conlist,
    constr,
    validator,
)


class Agent(BaseModel):
    """
    Agent artifact.  # noqa: E501
    """

    type: Optional[StrictStr] = "agent"
    name: constr(strict=True) = Field(
        default=..., description="Names cannot repeat accross agent +uiType=Hidden"
    )
    available_tools: conlist(constr(strict=True, min_length=1)) = Field(
        default=..., description="+sort=30 +uiType=Hidden"
    )
    goal: constr(strict=True, max_length=128, min_length=1) = Field(
        default=..., description="+sort=10 +uiType=TextArea"
    )
    instruction: constr(strict=True, max_length=2620, min_length=1) = Field(
        default=...,
        description='`instruction` is the system prompt for now. (2.5 * 1024) +sort=20 +uiType=AgentInstructions +uiProps={"helpText":"Use the syntax ${Tool FQN} to reference a tool, and ${AGENT FQN} to reference another agent"}',
    )
    model_id: constr(strict=True, min_length=1) = Field(
        default=...,
        description='+sort=40 +uiType=EnabledModelSelector +uiProps={"searchable":true,"modelType":"chat","providerType":"openai"}',
    )
    __properties = [
        "type",
        "name",
        "available_tools",
        "goal",
        "instruction",
        "model_id",
    ]

    @validator("type")
    def type_validate_enum(cls, value):
        """Validates the enum"""
        if value is None:
            return value

        if value not in ("agent",):
            raise ValueError("must be one of enum values ('agent')")
        return value

    @validator("name")
    def name_validate_regular_expression(cls, value):
        """Validates the regular expression"""
        if not re.match(r"^[a-zA-Z][a-zA-Z0-9\-]{1,30}[a-zA-Z0-9]$", value):
            raise ValueError(
                r"must validate the regular expression /^[a-zA-Z][a-zA-Z0-9\-]{1,30}[a-zA-Z0-9]$/"
            )
        return value

    class Config:
        """Pydantic configuration"""

        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Agent:
        """Create an instance of Agent from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True, exclude={}, exclude_none=True)
        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> Agent:
        """Create an instance of Agent from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return Agent.parse_obj(obj)

        _obj = Agent.parse_obj(
            {
                "type": obj.get("type") if obj.get("type") is not None else "agent",
                "name": obj.get("name"),
                "available_tools": obj.get("available_tools"),
                "goal": obj.get("goal"),
                "instruction": obj.get("instruction"),
                "model_id": obj.get("model_id"),
            }
        )
        return _obj



# generated by datamodel-codegen:
#   filename:  artifacts.json
#   timestamp: 2025-01-03T09:11:54+00:00

from __future__ import annotations

from enum import Enum
from typing import Any, Dict, List, Literal, Optional, Union

from truefoundry.pydantic_v1 import BaseModel, Field, conint, constr


class Agent(BaseModel):
    type: Literal["agent"] = Field(..., description="+value=agent")
    name: constr(regex=r"^[a-zA-Z][a-zA-Z0-9\-]{1,30}[a-zA-Z0-9]$") = Field(
        ..., description="Names cannot repeat accross agent\n+uiType=Hidden"
    )
    available_tools: List[constr(min_length=1)] = Field(
        ..., description="+sort=30\n+uiType=Hidden"
    )
    goal: constr(min_length=1, max_length=128) = Field(
        ..., description="+sort=10\n+uiType=TextArea"
    )
    instruction: constr(min_length=1, max_length=2620) = Field(
        ...,
        description='`instruction` is the system prompt for now. (2.5 * 1024)\n+sort=20\n+uiType=AgentInstructions\n+uiProps={"helpText":"Use the syntax ${Tool FQN} to reference a tool, and ${AGENT FQN} to reference another agent"}',
    )
    model_id: constr(min_length=1) = Field(
        ...,
        description='+sort=40\n+uiType=EnabledModelSelector\n+uiProps={"searchable":true,"modelType":"chat","providerType":"openai"}',
    )


class Method(str, Enum):
    """
    +sort=50
    +uiType=Hidden
    """

    get = "get"
    post = "post"
    put = "put"
    delete = "delete"
    patch = "patch"


class AgentWithFQN(Agent):
    id: str
    fqn: str


class BaseArtifactVersion(BaseModel):
    description: Optional[constr(max_length=512)] = Field(
        None,
        description="+label=Description\n+usage=Description of the artifact or model version\n+docs=Description of the artifact or model version",
    )
    metadata: Dict[str, Any] = Field(
        ...,
        description="+label=Metadata\n+docs=Metadata for the artifact or model version\n+usage=Metadata for the artifact or model version\n+uiType=JsonInput",
    )
    version_alias: Optional[
        constr(regex=r"^v[a-zA-Z0-9.-]*([a-zA-Z0-9]+)$", max_length=128)
    ] = Field(
        None,
        description="+label=Version Alias\n+usage=The version alias for artifact or model version which should start with 'v' followed by alphanumeric and it can include '.' and '-' in between (e.g. v1.0.0, v-prod, v-dev, etc)\n+docs=The version alias for artifact or model version which should start with 'v' followed by alphanumeric and it can include '.' and '-' in between (e.g. v1.0.0, v-prod, v-dev, etc)\n+message=The version alias should start with 'v' followed by alphanumeric and it can include '.' and '-' in between (e.g. v1.0.0, v-prod, v-dev, etc)",
    )


class MimeType(str, Enum):
    """
    +label=MIME Type
    +usage=MIME type of the content
    """

    text_plain = "text/plain"
    application_json = "application/json"
    image_png = "image/png"
    image_jpeg = "image/jpeg"
    application_x_directory = "application/x-directory"


class BlobStorageReference(BaseModel):
    """
    +usage=Blob Storage Location
    +label=Blob Storage Location
    +docs=Defines the structure for blob storage content references, including type, path, and mime_type
    """

    type: Literal["blob-storage"]
    path: constr(regex=r"^.{1,}$") = Field(
        ..., description="+label=Path\n+usage=Path to the content in blob storage"
    )
    mime_type: MimeType = Field(
        ..., description="+label=MIME Type\n+usage=MIME type of the content"
    )


class ExternalArtifactSource(BaseModel):
    """
    +label=External
    """

    type: Literal["external"] = Field(
        ...,
        description="+label=Type\n+usage=Type of the source\n+value=external\n+type=External",
    )
    uri: str = Field(..., description="+label=URI\n+usage=URI of the Artifact source")


class FastAIFramework(BaseModel):
    """
    +docs=FastAI framework for the model version
    +label=FastAI
    """

    type: Literal["fastai"] = Field(
        ..., description="+label=Type\n+usage=Type of the framework\n+value=fastai"
    )


class GluonFramework(BaseModel):
    """
    +docs=Gluon framework for the model version
    +label=Gluon
    """

    type: Literal["gluon"] = Field(
        ..., description="+label=Type\n+usage=Type of the framework\n+value=gluon"
    )


class H2OFramework(BaseModel):
    """
    +docs=H2O framework for the model version
    +label=H2O
    """

    type: Literal["h2o"] = Field(
        ..., description="+label=Type\n+usage=Type of the framework\n+value=h2o"
    )


class ImageUrl(BaseModel):
    """
    +label=URL for the image
    +usage=The URL for the image, must be a non-empty valid HTTPS URL or a data URL
    """

    url: Union[
        constr(
            regex=r"^\b((https?://)?(?:www\.)?[-a-zA-Z0-9@:%._\+~#=]{1,256}\.[a-zA-Z0-9()]{1,6}(?:[-a-zA-Z0-9()@:%_\+.~#?&/=]*)|data:image/[a-zA-Z]+;base64,[a-zA-Z0-9+/=\s]+)$"
        ),
        BlobStorageReference,
    ] = Field(
        ...,
        description="+label=URL for the image\n+usage=The URL for the image, must be a non-empty valid HTTPS URL or a data URL",
    )


class ImageContentPart(BaseModel):
    """
    +usage=Image URL
    +label=Image URL
    """

    type: Literal["image_url"]
    image_url: ImageUrl = Field(
        ...,
        description="+label=URL for the image\n+usage=The URL for the image, must be a non-empty valid HTTPS URL or a data URL",
    )


class KerasFramework(BaseModel):
    """
    +docs=Keras framework for the model version
    +label=Keras
    """

    type: Literal["keras"] = Field(
        ..., description="+label=Type\n+usage=Type of the framework\n+value=keras"
    )


class LightGBMFramework(BaseModel):
    """
    +docs=LightGBM framework for the model version
    +label=LightGBM
    """

    type: Literal["lightgbm"] = Field(
        ..., description="+label=Type\n+usage=Type of the framework\n+value=lightgbm"
    )


class Parameters(BaseModel):
    """
    +usage=Parameters for the provider
    +label=Parameters
    +docs=Key-value pairs to store additional parameters for the provider
    """

    max_tokens: Optional[int] = None
    temperature: Optional[float] = None
    top_k: Optional[float] = None
    top_p: Optional[float] = None
    stop: Optional[Union[List[str], str]] = None


class ModelConfiguration(BaseModel):
    """
    +label=Model Configuration
    +icon=fa-cogs:#326ce5
    +message=Configuration details for the provider and corresponding config
    +usage=Define the provider and its configuration
    +docs=Configuration settings specific to the provider, including model settings and other parameters.
    """

    provider: str = Field(
        ...,
        description='+label=Provider Name\n+usage=Name of the provider, must be non-empty, e.g., "openai", "google_gemini"',
    )
    model: constr(regex=r"^[a-zA-Z][a-zA-Z0-9\-]{1,30}[a-zA-Z0-9]$") = Field(
        ..., description="+label=Model name\n+usage=Model name, must be non-empty"
    )
    parameters: Optional[Parameters] = Field(
        None,
        description="+usage=Parameters for the provider\n+label=Parameters\n+docs=Key-value pairs to store additional parameters for the provider",
    )
    extra_parameters: Optional[Dict[str, Any]] = Field(
        None,
        description="+usage=Extra parameters for the provider\n+label=Extra Parameters\n+docs=Additional parameters for the provider",
    )


class ModelVersionEnvironment(BaseModel):
    """
    +label=Environment
    """

    python_version: Optional[constr(regex=r"^\d+(\.\d+){1,2}([\-\.a-z0-9]+)?$")] = (
        Field(
            None,
            description="+label=Python Version\n+usage=Python version for the model version",
        )
    )
    pip_packages: Optional[List[str]] = Field(
        None,
        description="+label=PIP Packages\n+usage=PIP packages for the model version",
    )


class ONNXFramework(BaseModel):
    """
    +docs=ONNX framework for the model version
    +label=ONNX
    """

    type: Literal["onnx"] = Field(
        ..., description="+label=Type\n+usage=Type of the framework\n+value=onnx"
    )


class PaddleFramework(BaseModel):
    """
    +docs=PaddlePaddle framework for the model version
    +label=Paddle
    """

    type: Literal["paddle"] = Field(
        ..., description="+label=Type\n+usage=Type of the framework\n+value=paddle"
    )


class PyTorchFramework(BaseModel):
    """
    +docs=PyTorch framework for the model version
    +label=PyTorch
    """

    type: Literal["pytorch"] = Field(
        ..., description="+label=Type\n+usage=Type of the framework\n+value=pytorch"
    )


class InferMethodName(str, Enum):
    """
    +label=Inference Method Name
    +usage=Name of the method used for inference
    """

    predict = "predict"
    predict_proba = "predict_proba"


class SklearnModelSchema(BaseModel):
    """
    +label=Sklearn Model Schema
    """

    infer_method_name: InferMethodName = Field(
        ...,
        description="+label=Inference Method Name\n+usage=Name of the method used for inference",
    )
    inputs: List[Dict[str, Any]] = Field(
        ..., description="+label= Input Schema\n+usage=Schema of the input"
    )
    outputs: List[Dict[str, Any]] = Field(
        ..., description="+label= Output Schema\n+usage=Schema of the output"
    )


class SklearnSerializationFormat(str, Enum):
    """
    +label=Serialization format
    +usage=Serialization format used for sklearn models
    """

    cloudpickle = "cloudpickle"
    joblib = "joblib"
    pickle = "pickle"


class SpaCyFramework(BaseModel):
    """
    +docs=spaCy framework for the model version
    +label=SpaCy
    """

    type: Literal["spacy"] = Field(
        ..., description="+label=Type\n+usage=Type of the framework\n+value=spacy"
    )


class StatsModelsFramework(BaseModel):
    """
    +docs=StatsModels framework for the model version
    +label=StatsModels
    """

    type: Literal["statsmodels"] = Field(
        ..., description="+label=Type\n+usage=Type of the framework\n+value=statsmodels"
    )


class SystemMessage(BaseModel):
    """
    +usage=System message
    +docs=Defines the structure of a system message, including role and content
    +label=System Message
    """

    role: Literal["system"]
    content: Union[constr(regex=r"^.[\s\S]*$"), BlobStorageReference] = Field(
        ..., description="+label=Content\n+usage=Text content for the system message"
    )
    name: Optional[str] = Field(
        None, description="+label=Name\n+usage=Name of the system"
    )


class TensorFlowFramework(BaseModel):
    """
    +docs=TensorFlow framework for the model version
    +label=TensorFlow
    """

    type: Literal["tensorflow"] = Field(
        ..., description="+label=Type\n+usage=Type of the framework\n+value=tensorflow"
    )


class TextContentPart(BaseModel):
    """
    +usage=Text content
    +label=Text content
    """

    type: Literal["text"]
    text: Union[constr(regex=r"^.[\s\S]*$"), BlobStorageReference]


class LibraryName(str, Enum):
    """
    +label=Library Name
    +usage=Name of the library for the framework
    """

    transformers = "transformers"
    sentence_transformers = "sentence-transformers"
    diffusers = "diffusers"


class TransformersFramework(BaseModel):
    """
    +docs=Transformers framework for the model version
    +label=Transformers
    +value=transformers
    """

    type: Literal["transformers"] = Field(
        ...,
        description="+label=Type\n+usage=Type of the framework\n+value=transformers",
    )
    library_name: Optional[LibraryName] = Field(
        "transformers",
        description="+label=Library Name\n+usage=Name of the library for the framework",
    )
    pipeline_tag: Optional[str] = Field(
        None,
        description="+label=Pipeline Tag\n+usage=The `pipeline()` task this model can be used with e.g. `text-generation`. See [huggingface docs](https://huggingface.co/docs/transformers/main/en/main_classes/pipelines#transformers.pipeline.task) for all possible values\n+docs=Pipeline tag for the framework",
    )
    base_model: Optional[str] = Field(
        None,
        description="+label=Base Model\n+usage=Base model Id. If this is a finetuned model, this points to the base model used for finetuning\n+docs=Base model Id. If this is a finetuned model, this points to the base model used for finetuning",
    )


class TrueFoundryArtifactSource(BaseModel):
    """
    +usage=Source for the Artifact or Model
    +label=Upload
    """

    type: Literal["truefoundry"] = Field(
        ...,
        description="+usage=Type of the source\n+type=Upload\n+value=truefoundry\n+label=Upload",
    )
    uri: str = Field(
        ...,
        description='+label=URI\n+uiType=ModelUploadInput\n+uiProps={"hideClear":true}',
    )


class UserMessage(BaseModel):
    """
    +usage=User message
    +docs=Defines the structure of a user message, including role and content
    +label=User Message
    """

    role: Literal["user"]
    content: Union[
        constr(regex=r"^.[\s\S]*$"),
        BlobStorageReference,
        List[Union[TextContentPart, ImageContentPart]],
    ] = Field(
        ..., description="+label=Content\n+usage=Text content for the user message"
    )
    name: Optional[str] = Field(
        None, description="+label=Name\n+usage=Name of the user message"
    )


class XGBoostModelSchema(BaseModel):
    """
    +label=XGBoost Model Schema
    """

    infer_method_name: Literal["predict"] = Field(
        ...,
        description="+label=Inference Method Name\n+usage=Name of the method used for inference",
    )
    inputs: List[Dict[str, Any]] = Field(
        ..., description="+label= Input Schema\n+usage=Schema of the input"
    )
    outputs: List[Dict[str, Any]] = Field(
        ..., description="+label= Output Schema\n+usage=Schema of the output"
    )


class XGBoostSerializationFormat(str, Enum):
    """
    +label=Serialization format
    +usage=Serialization format used for XGBoost models
    """

    cloudpickle = "cloudpickle"
    joblib = "joblib"
    pickle = "pickle"
    json = "json"


class AgentOpenAPITool(BaseModel):
    type: Literal["openapi-tool"] = Field(..., description="+value=openapi-tool")
    name: constr(regex=r"^[a-zA-Z][a-zA-Z0-9\-]{1,30}[a-zA-Z0-9]$") = Field(
        ..., description="Names cannot repeat accross tools\n+uiType=Hidden"
    )
    description: constr(min_length=1, max_length=128) = Field(
        ..., description="+sort=10\n+uiType=TextArea"
    )
    openapi_spec: Union[BlobStorageReference, Dict[str, Any]] = Field(
        ..., description="+sort=20\n+uiType=OpenapiSchema"
    )
    base_url: constr(
        regex=r"^(https?://)?(?:www\.)?[-a-zA-Z0-9@:%._\+~#=]{1,256}\.[a-zA-Z0-9()]{1,6}(?:[-a-zA-Z0-9()@:%_\+.~#?&/=]*)$"
    ) = Field(..., description="+sort=30")
    path: constr(min_length=1) = Field(
        ...,
        description="+sort=40\n+uiType=MethodPathSelector\n+label=Method + Path\n+usage=Select one Method and Path for this tool",
    )
    method: Method = Field(..., description="+sort=50\n+uiType=Hidden")
    headers: Optional[Dict[str, str]] = Field(
        None,
        description='+sort=60\n+uiType=KV\n+uiProps={"allowSecrets":true,"secretConfig":{"enableNew":true,"hideOptions":true}}',
    )


class AgentOpenAPIToolWithFQN(AgentOpenAPITool):
    id: str
    fqn: str


class ArtifactVersion(BaseArtifactVersion):
    type: Literal["artifact-version"] = Field(
        ...,
        description='+label=Type\n+usage=Artifact Version\n+value="artifact-version"',
    )
    source: Union[TrueFoundryArtifactSource, ExternalArtifactSource] = Field(
        ...,
        description="+label=Artifact Source\n+usage=Source for the Artifact version\n+uiType=Group",
    )
    step: conint(ge=0) = Field(0, description="+label=Step")


class AssistantMessage(BaseModel):
    """
    +usage=Assistant message
    +docs=Defines the structure of an assistant message, including role and content
    +label=Assistant Message
    """

    role: Literal["assistant"]
    content: Union[constr(regex=r"^.[\s\S]*$"), BlobStorageReference] = Field(
        ..., description="+label=Content\n+usage=Text content for the assistant message"
    )
    name: Optional[str] = Field(
        None, description="+label=Name\n+usage=Name of the assistant message"
    )


class BasePrompt(BaseModel):
    """
    Main prompt structure
    """

    name: constr(regex=r"^[a-zA-Z][a-zA-Z0-9-_]{0,254}[a-zA-Z0-9]$") = Field(
        ...,
        description="+label=Prompt Name\n+icon=fa-desktop:#326ce5\n+message=Alphanumeric word, may contain '-' with a maximum length of 256 characters\n+sort=1000\n+usage=Name of the Prompt. This uniquely identifies it in the workspace.\n+docs=The unique name for the prompt, consisting of alphanumeric characters and dashes, max length 100.",
    )
    description: Optional[constr(min_length=1, max_length=512)] = Field(
        None,
        description="+label=Description\n+icon=fa-desktop:#326ce5\n+message=Description in a maximum of 512 characters\n+sort=2000\n+usage=Description of the prompt\n+docs=An optional description for the prompt, with a maximum length of 512 characters.\n+optional",
    )
    model_configuration: ModelConfiguration
    metadata: Dict[str, str] = Field(
        ...,
        description="+label=Metadata\n+icon=fa-info-circle:#326ce5\n+message=Additional metadata for the prompt\n+sort=7000\n+usage=Define additional metadata for the prompt\n+docs=Key-value pairs to store additional metadata related to the prompt.",
    )


class ChatPrompt(BasePrompt):
    type: Literal["chat_prompt"] = Field(
        ...,
        description='+label=Type\n+icon=fa-desktop:#326ce5\n+message=Type of the prompt\n+sort=3000\n+usage=Type of the prompt\n+docs=Type of the prompt, set to "chat_prompt"\n+default="chat_prompt"',
    )
    messages: List[Union[SystemMessage, AssistantMessage, UserMessage]] = Field(
        ...,
        description="+sort=4000\n+usage=Chat completion messages\n+label=Messages in the chat conversation\n+message=Chat completion messages\n+usage=List of messages in the chat conversation, must be non-empty\n+docs=Messages that define the chat conversation, including system, assistant, and user messages.",
    )
    variables: Optional[Dict[str, Optional[str]]] = Field(
        None,
        description="+label=Variables\n+usage=Variables for the chat completion messages to be used in the prompt messages\n+sort=5000",
    )


class SklearnFramework(BaseModel):
    """
    +docs=Scikit-learn framework for the model version
    +label=Sklearn
    """

    type: Literal["sklearn"] = Field(
        ..., description="+label=Type\n+usage=Type of the framework\n+value=sklearn"
    )
    model_filepath: Optional[str] = Field(
        None,
        description="+label=Model file path\n+usage=Relative path to the model file",
    )
    serialization_format: Optional[SklearnSerializationFormat] = None
    model_schema: Optional[SklearnModelSchema] = None


class XGBoostFramework(BaseModel):
    """
    +docs=XGBoost framework for the model version
    +label=XGBoost
    """

    type: Literal["xgboost"] = Field(
        ..., description="+label=Type\n+usage=Type of the framework\n+value=xgboost"
    )
    serialization_format: Optional[XGBoostSerializationFormat] = None
    model_filepath: Optional[str] = Field(
        None,
        description="+label=Model file path\n+usage=Relative path to the model file",
    )
    model_schema: Optional[XGBoostModelSchema] = None


class AgentApp(BaseModel):
    type: Literal["agent-app"] = Field(..., description="+value=agent-app")
    tools: List[AgentOpenAPIToolWithFQN]
    agents: List[AgentWithFQN]
    root_agent: constr(min_length=1)


class ModelVersion(BaseArtifactVersion):
    type: Literal["model-version"] = Field(
        ..., description='+label=Type\n+usage=Model Version\n+value="model-version"'
    )
    source: Union[TrueFoundryArtifactSource, ExternalArtifactSource] = Field(
        ...,
        description="+label=Model Source\n+usage=Source for the model version\n+uiType=Group",
    )
    framework: Optional[
        Union[
            TransformersFramework,
            TensorFlowFramework,
            SklearnFramework,
            PyTorchFramework,
            KerasFramework,
            XGBoostFramework,
            LightGBMFramework,
            FastAIFramework,
            H2OFramework,
            ONNXFramework,
            SpaCyFramework,
            StatsModelsFramework,
            GluonFramework,
            PaddleFramework,
        ]
    ] = Field(
        None, description="+label=Framework\n+usage=Framework for the model version"
    )
    environment: Optional[ModelVersionEnvironment] = None
    step: conint(ge=0) = Field(0, description="+label=Step")


class VersionedArtifactType(BaseModel):
    __root__: Union[
        ChatPrompt, AgentOpenAPITool, Agent, AgentApp, ModelVersion, ArtifactVersion
    ]



from truefoundry.common.warnings import (
    suppress_truefoundry_deprecation_warnings,
    surface_truefoundry_deprecation_warnings,
)
from truefoundry.deploy.core import login, logout

surface_truefoundry_deprecation_warnings()
__all__ = [
    "login",
    "logout",
    "suppress_truefoundry_deprecation_warnings",
]



# coding: utf-8

"""
FastAPI

No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

The version of the OpenAPI document: 0.1.0
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

import json
import re  # noqa: F401

from aenum import Enum


class SklearnSerializationFormat(str, Enum):
    """
    +label=Serialization format +usage=Serialization format used for sklearn models
    """

    """
    allowed enum values
    """
    CLOUDPICKLE = "cloudpickle"
    JOBLIB = "joblib"
    PICKLE = "pickle"

    @classmethod
    def from_json(cls, json_str: str) -> SklearnSerializationFormat:
        """Create an instance of SklearnSerializationFormat from a JSON string"""
        return SklearnSerializationFormat(json.loads(json_str))



# coding: utf-8

"""
FastAPI

No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

The version of the OpenAPI document: 0.1.0
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations

import json
import pprint
import re  # noqa: F401
from typing import Optional, Union

from truefoundry.pydantic_v1 import BaseModel, Field, StrictFloat, StrictInt, StrictStr


class LogMetricRequestDto(BaseModel):
    """
    LogMetricRequestDto
    """

    run_id: Optional[StrictStr] = None
    run_uuid: Optional[StrictStr] = None
    key: StrictStr = Field(...)
    value: Union[StrictFloat, StrictInt] = Field(...)
    timestamp: StrictInt = Field(...)
    step: Optional[StrictInt] = 0
    __properties = ["run_id", "run_uuid", "key", "value", "timestamp", "step"]

    class Config:
        """Pydantic configuration"""

        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> LogMetricRequestDto:
        """Create an instance of LogMetricRequestDto from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True, exclude={}, exclude_none=True)
        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> LogMetricRequestDto:
        """Create an instance of LogMetricRequestDto from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return LogMetricRequestDto.parse_obj(obj)

        _obj = LogMetricRequestDto.parse_obj(
            {
                "run_id": obj.get("run_id"),
                "run_uuid": obj.get("run_uuid"),
                "key": obj.get("key"),
                "value": obj.get("value"),
                "timestamp": obj.get("timestamp"),
                "step": obj.get("step") if obj.get("step") is not None else 0,
            }
        )
        return _obj



# coding: utf-8

"""
FastAPI

No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

The version of the OpenAPI document: 0.1.0
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations

import json
import pprint
import re  # noqa: F401
from typing import Dict, Optional

from truefoundry.ml.autogen.client.models.method import Method
from truefoundry.ml.autogen.client.models.openapi_spec import OpenapiSpec
from truefoundry.pydantic_v1 import BaseModel, Field, StrictStr, constr, validator


class AgentOpenAPITool(BaseModel):
    """
    OpenAPI Tool artifact.  # noqa: E501
    """

    type: Optional[StrictStr] = "openapi-tool"
    name: constr(strict=True) = Field(
        default=..., description="Names cannot repeat accross tools +uiType=Hidden"
    )
    description: constr(strict=True, max_length=128, min_length=1) = Field(
        default=..., description="+sort=10 +uiType=TextArea"
    )
    openapi_spec: OpenapiSpec = Field(...)
    base_url: constr(strict=True) = Field(default=..., description="+sort=30")
    path: constr(strict=True, min_length=1) = Field(
        default=...,
        description="+sort=40 +uiType=MethodPathSelector +label=Method + Path +usage=Select one Method and Path for this tool",
    )
    method: Method = Field(default=..., description="+sort=50 +uiType=Hidden")
    headers: Optional[Dict[str, StrictStr]] = Field(
        default=None,
        description='+sort=60 +uiType=KV +uiProps={"allowSecrets":true,"secretConfig":{"enableNew":true,"hideOptions":true}}',
    )
    __properties = [
        "type",
        "name",
        "description",
        "openapi_spec",
        "base_url",
        "path",
        "method",
        "headers",
    ]

    @validator("type")
    def type_validate_enum(cls, value):
        """Validates the enum"""
        if value is None:
            return value

        if value not in ("openapi-tool",):
            raise ValueError("must be one of enum values ('openapi-tool')")
        return value

    @validator("name")
    def name_validate_regular_expression(cls, value):
        """Validates the regular expression"""
        if not re.match(r"^[a-zA-Z][a-zA-Z0-9\-]{1,30}[a-zA-Z0-9]$", value):
            raise ValueError(
                r"must validate the regular expression /^[a-zA-Z][a-zA-Z0-9\-]{1,30}[a-zA-Z0-9]$/"
            )
        return value

    @validator("base_url")
    def base_url_validate_regular_expression(cls, value):
        """Validates the regular expression"""
        if not re.match(
            r"^(https?:\/\/)?(?:www\.)?[-a-zA-Z0-9@:%._\+~#=]{1,256}\.[a-zA-Z0-9()]{1,6}(?:[-a-zA-Z0-9()@:%_\+.~#?&\/=]*)$",
            value,
        ):
            raise ValueError(
                r"must validate the regular expression /^(https?:\/\/)?(?:www\.)?[-a-zA-Z0-9@:%._\+~#=]{1,256}\.[a-zA-Z0-9()]{1,6}(?:[-a-zA-Z0-9()@:%_\+.~#?&\/=]*)$/"
            )
        return value

    class Config:
        """Pydantic configuration"""

        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> AgentOpenAPITool:
        """Create an instance of AgentOpenAPITool from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True, exclude={}, exclude_none=True)
        # override the default output from truefoundry.pydantic_v1 by calling `to_dict()` of openapi_spec
        if self.openapi_spec:
            _dict["openapi_spec"] = self.openapi_spec.to_dict()
        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> AgentOpenAPITool:
        """Create an instance of AgentOpenAPITool from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return AgentOpenAPITool.parse_obj(obj)

        _obj = AgentOpenAPITool.parse_obj(
            {
                "type": obj.get("type")
                if obj.get("type") is not None
                else "openapi-tool",
                "name": obj.get("name"),
                "description": obj.get("description"),
                "openapi_spec": OpenapiSpec.from_dict(obj.get("openapi_spec"))
                if obj.get("openapi_spec") is not None
                else None,
                "base_url": obj.get("base_url"),
                "path": obj.get("path"),
                "method": obj.get("method"),
                "headers": obj.get("headers"),
            }
        )
        return _obj



from typing import Literal, Union

from truefoundry.deploy.auto_gen import models
from truefoundry.deploy.lib.model.entity import Deployment
from truefoundry.deploy.v2.lib.deploy import deploy_component
from truefoundry.deploy.v2.lib.patched_models import LocalSource
from truefoundry.pydantic_v1 import BaseModel, Field, conint


class DeployablePatchedModelBase(BaseModel):
    class Config:
        extra = "forbid"

    def deploy(self, workspace_fqn: str, wait: bool = True) -> Deployment:
        return deploy_component(component=self, workspace_fqn=workspace_fqn, wait=wait)


class Service(models.Service, DeployablePatchedModelBase):
    type: Literal["service"] = "service"
    resources: models.Resources = Field(default_factory=models.Resources)
    # This is being patched because cue export marks this as a "number"
    replicas: Union[conint(ge=0, le=100), models.ServiceAutoscaling] = Field(
        1,
        description="+label=Replicas\n+usage=Replicas of service you want to run\n+icon=fa-clone\n+sort=3",
    )


class Job(models.Job, DeployablePatchedModelBase):
    type: Literal["job"] = "job"
    resources: models.Resources = Field(default_factory=models.Resources)


class Notebook(models.Notebook, DeployablePatchedModelBase):
    type: Literal["notebook"] = "notebook"
    resources: models.Resources = Field(default_factory=models.Resources)


class Codeserver(models.Codeserver, DeployablePatchedModelBase):
    type: Literal["codeserver"] = "codeserver"


class Helm(models.Helm, DeployablePatchedModelBase):
    type: Literal["helm"] = "helm"


class Volume(models.Volume, DeployablePatchedModelBase):
    type: Literal["volume"] = "volume"


class ApplicationSet(models.ApplicationSet, DeployablePatchedModelBase):
    type: Literal["application-set"] = "application-set"


class AsyncService(models.AsyncService, DeployablePatchedModelBase):
    type: Literal["async-service"] = "async-service"
    replicas: Union[conint(ge=0, le=100), models.AsyncServiceAutoscaling] = 1
    resources: models.Resources = Field(default_factory=models.Resources)


class SSHServer(models.SSHServer, DeployablePatchedModelBase):
    type: Literal["ssh-server"] = "ssh-server"
    resources: models.Resources = Field(default_factory=models.Resources)


class Application(models.Application, DeployablePatchedModelBase):
    def deploy(self, workspace_fqn: str, wait: bool = True) -> Deployment:
        if isinstance(self.__root__, models.Workflow):
            from truefoundry.deploy.v2.lib.deploy_workflow import deploy_workflow

            return deploy_workflow(
                workflow=self.__root__,
                workspace_fqn=workspace_fqn,
                wait=wait,
            )
        else:
            return deploy_component(
                component=self.__root__, workspace_fqn=workspace_fqn, wait=wait
            )


class Workflow(models.Workflow, DeployablePatchedModelBase):
    type: Literal["workflow"] = "workflow"
    source: Union[models.RemoteSource, models.LocalSource] = Field(
        default_factory=lambda: LocalSource(local_build=False)
    )

    def deploy(self, workspace_fqn: str, wait: bool = True) -> Deployment:
        from truefoundry.deploy.v2.lib.deploy_workflow import deploy_workflow

        return deploy_workflow(workflow=self, workspace_fqn=workspace_fqn, wait=wait)
